# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       url
"""
from typing import List, Tuple, Optional

from .exception import HttpProtocolException
from ..common.utils import text_
from ..common.constants import AT, COLON, SLASH, DEFAULT_ALLOWED_URL_SCHEMES


class Url:
    """``urllib.urlparse`` doesn't work for proxy.py, so we wrote a simple URL.

    Currently, URL only implements what is necessary for HttpParser to work.
    """

    def __init__(
            self,
            scheme: Optional[bytes] = None,
            username: Optional[bytes] = None,
            password: Optional[bytes] = None,
            hostname: Optional[bytes] = None,
            port: Optional[int] = None,
            remainder: Optional[bytes] = None,
    ) -> None:
        self.scheme: Optional[bytes] = scheme
        self.username: Optional[bytes] = username
        self.password: Optional[bytes] = password
        self.hostname: Optional[bytes] = hostname
        self.port: Optional[int] = port
        self.remainder: Optional[bytes] = remainder

    @property
    def has_credentials(self) -> bool:
        """Returns true if both username and password components are present."""
        return self.username is not None and self.password is not None

    def __str__(self) -> str:
        url = ''
        if self.scheme:
            url += '{0}://'.format(text_(self.scheme))
        if self.hostname:
            url += text_(self.hostname)
        if self.port:
            url += ':{0}'.format(self.port)
        if self.remainder:
            url += text_(self.remainder)
        return url

    @classmethod
    def from_bytes(cls, raw: bytes, allowed_url_schemes: Optional[List[bytes]] = None) -> 'Url':
        """A URL within proxy.py core can have several styles,
        because proxy.py supports both proxy and web server use cases.

        Example:
        For a Web server, url is like ``/`` or ``/get`` or ``/get?key=value``
        For a HTTPS connect tunnel, url is like ``httpbin.org:443``
        For a HTTP proxy request, url is like ``http://httpbin.org/get``

        proxy.py internally never expects a https scheme in the request line.
        But `Url` class provides support for parsing any scheme present in the URLs.
        e.g. ftp, icap etc.

        If a url with no scheme is parsed, e.g. ``//host/abc.js``, then scheme
        defaults to `http`.

        Further:
        1) URL may contain unicode characters
        2) URL may contain IPv4 and IPv6 format addresses instead of domain names
        """
        # SLASH == 47, check if URL starts with single slash but not double slash
        starts_with_single_slash = raw[0] == 47
        starts_with_double_slash = starts_with_single_slash and \
            len(raw) >= 2 and \
            raw[1] == 47
        if starts_with_single_slash and \
                not starts_with_double_slash:
            return cls(remainder=raw)
        scheme = None
        rest = None
        if not starts_with_double_slash:
            # Find scheme
            parts = raw.split(b'://', 1)
            if len(parts) == 2:
                scheme = parts[0]
                rest = parts[1]
                if scheme not in (allowed_url_schemes or DEFAULT_ALLOWED_URL_SCHEMES):
                    raise HttpProtocolException(
                        'Invalid scheme received in the request line %r' % raw,
                    )
        else:
            rest = raw[len(SLASH + SLASH):]
        if scheme is not None or starts_with_double_slash:
            assert rest is not None
            parts = rest.split(SLASH, 1)
            username, password, host, port = Url._parse(parts[0])
            return cls(
                scheme=scheme if not starts_with_double_slash else b'http',
                username=username,
                password=password,
                hostname=host,
                port=port,
                remainder=None if len(parts) == 1 else (
                    SLASH + parts[1]
                ),
            )
        username, password, host, port = Url._parse(raw)
        return cls(username=username, password=password, hostname=host, port=port)

    @staticmethod
    def _parse(raw: bytes) -> Tuple[
            Optional[bytes],
            Optional[bytes],
            bytes,
            Optional[int],
    ]:
        split_at = raw.split(AT, 1)
        username, password = None, None
        if len(split_at) == 2:
            username, password = split_at[0].split(COLON)
        parts = split_at[-1].split(COLON, 2)
        num_parts = len(parts)
        port: Optional[int] = None
        # No port found
        if num_parts == 1:
            return username, password, parts[0], None
        # Host and port found
        if num_parts == 2:
            return username, password, COLON.join(parts[:-1]), int(parts[-1])
        # More than a single COLON i.e. IPv6 scenario
        try:
            # Try to resolve last part as an int port
            last_token = parts[-1].split(COLON)
            port = int(last_token[-1])
            host = COLON.join(parts[:-1]) + COLON + \
                COLON.join(last_token[:-1])
        except ValueError:
            # If unable to convert last part into port,
            # treat entire data as host
            host, port = raw, None
        # patch up invalid ipv6 scenario
        rhost = host.decode('utf-8')
        if COLON.decode('utf-8') in rhost and \
                rhost[0] != '[' and \
                rhost[-1] != ']':
            host = b'[' + host + b']'
        return username, password, host, port



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from .names import EventNames, eventNames
from .queue import EventQueue
from .manager import EventManager
from .dispatcher import EventDispatcher
from .subscriber import EventSubscriber


__all__ = [
    'eventNames',
    'EventNames',
    'EventQueue',
    'EventDispatcher',
    'EventSubscriber',
    'EventManager',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import ssl
import time
import errno
import socket
import asyncio
import logging
import selectors
from typing import Any, List, Type, Tuple, Optional

from .parser import HttpParser, httpParserTypes, httpParserStates
from .plugin import HttpProtocolHandlerPlugin
from .exception import HttpProtocolException
from .protocols import httpProtocols
from .responses import BAD_REQUEST_RESPONSE_PKT
from ..core.base import BaseTcpServerHandler
from .connection import HttpClientConnection
from ..common.types import Readables, Writables, SelectableEvents
from ..common.constants import DEFAULT_SELECTOR_SELECT_TIMEOUT


logger = logging.getLogger(__name__)


class HttpProtocolHandler(BaseTcpServerHandler[HttpClientConnection]):
    """HTTP, HTTPS, HTTP2, WebSockets protocol handler.

    Accepts `Client` connection and delegates to HttpProtocolHandlerPlugin.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.start_time: float = time.time()
        self.last_activity: float = self.start_time
        self.request: HttpParser = HttpParser(
            httpParserTypes.REQUEST_PARSER,
            enable_proxy_protocol=self.flags.enable_proxy_protocol,
        )
        self.selector: Optional[selectors.DefaultSelector] = None
        if not self.flags.threadless:
            self.selector = selectors.DefaultSelector()
        self.plugin: Optional[HttpProtocolHandlerPlugin] = None
        self.writes_teared: bool = False
        self.reads_teared: bool = False

    ##
    # initialize, is_inactive, shutdown, get_events, handle_events
    # overrides Work class definitions.
    ##

    @staticmethod
    def create(*args: Any) -> HttpClientConnection:  # pragma: no cover
        return HttpClientConnection(*args)

    def initialize(self) -> None:
        super().initialize()
        if self._encryption_enabled():
            self.work = HttpClientConnection(
                conn=self.work.connection,
                addr=self.work.addr,
            )

    def is_inactive(self) -> bool:
        if not self.work.has_buffer() and \
                self._connection_inactive_for() > self.flags.timeout:
            return True
        return False

    def shutdown(self) -> None:
        try:
            # Flush pending buffer in threaded mode only.
            #
            # For threadless mode, BaseTcpServerHandler implements
            # the must_flush_before_shutdown logic automagically.
            if self.selector and self.work.has_buffer():
                self._flush()
            # Invoke plugin.on_client_connection_close
            if self.plugin:
                self.plugin.on_client_connection_close()
            logger.debug(
                'Closing client connection %s has buffer %s' %
                (self.work.address, self.work.has_buffer()),
            )
            conn = self.work.connection
            # Unwrap if wrapped before shutdown.
            if self._encryption_enabled() and \
                    isinstance(self.work.connection, ssl.SSLSocket):
                conn = self.work.connection.unwrap()
            conn.shutdown(socket.SHUT_WR)
            logger.debug('Client connection shutdown successful')
        except OSError:
            pass
        finally:
            # Section 4.2.2.13 of RFC 1122 tells us that a close() with any pending readable data
            # could lead to an immediate reset being sent.
            #
            #   "A host MAY implement a 'half-duplex' TCP close sequence, so that an application
            #   that has called CLOSE cannot continue to read data from the connection.
            #   If such a host issues a CLOSE call while received data is still pending in TCP,
            #   or if new data is received after CLOSE is called, its TCP SHOULD send a RST to
            #   show that data was lost."
            #
            self.work.connection.close()
            logger.debug('Client connection closed')
            super().shutdown()

    async def get_events(self) -> SelectableEvents:
        # Get default client events
        events: SelectableEvents = await super().get_events()
        # HttpProtocolHandlerPlugin.get_descriptors
        if self.plugin:
            plugin_read_desc, plugin_write_desc = await self.plugin.get_descriptors()
            for rfileno in plugin_read_desc:
                if rfileno not in events:
                    events[rfileno] = selectors.EVENT_READ
                else:
                    events[rfileno] |= selectors.EVENT_READ
            for wfileno in plugin_write_desc:
                if wfileno not in events:
                    events[wfileno] = selectors.EVENT_WRITE
                else:
                    events[wfileno] |= selectors.EVENT_WRITE
        return events

    # We override super().handle_events and never call it
    async def handle_events(
            self,
            readables: Readables,
            writables: Writables,
    ) -> bool:
        """Returns True if proxy must tear down."""
        # Flush buffer for ready to write sockets
        self.writes_teared = await self.handle_writables(writables)
        if self.writes_teared:
            return True
        # Invoke plugin.write_to_descriptors
        if self.plugin:
            self.writes_teared = await self.plugin.write_to_descriptors(writables)
            if self.writes_teared:
                return True
        # Read from ready to read sockets if reads have not already teared down
        if not self.reads_teared:
            self.reads_teared = await self.handle_readables(readables)
            if not self.reads_teared:
                # Invoke plugin.read_from_descriptors
                if self.plugin:
                    self.reads_teared = await self.plugin.read_from_descriptors(
                        readables,
                    )
        # Wait until client buffer has flushed when reads has teared down but we can still write
        if self.reads_teared and not self.work.has_buffer():
            return True
        return False

    def handle_data(self, data: memoryview) -> Optional[bool]:
        """Handles incoming data from client."""
        if data is None:
            logger.debug('Client closed connection, tearing down...')
            self.work.closed = True
            return True
        try:
            # We don't parse incoming data any further after 1st HTTP request packet.
            #
            # Plugins can utilize on_client_data for such cases and
            # apply custom logic to handle request data sent after 1st
            # valid request.
            if self.request.state != httpParserStates.COMPLETE:
                if self._parse_first_request(data):
                    return True
            # HttpProtocolHandlerPlugin.on_client_data
            # Can raise HttpProtocolException to tear down the connection
            elif self.plugin:
                self.plugin.on_client_data(data)
        except HttpProtocolException as e:
            logger.warning('HttpProtocolException: %s' % e)
            response: Optional[memoryview] = e.response(self.request)
            if response:
                self.work.queue(response)
            return True
        return False

    async def handle_writables(self, writables: Writables) -> bool:
        if self.work.connection.fileno() in writables and self.work.has_buffer():
            logger.debug('Client is write ready, flushing...')
            self.last_activity = time.time()
            # TODO(abhinavsingh): This hook could just reside within server recv block
            # instead of invoking when flushed to client.
            #
            # Invoke plugin.on_response_chunk
            chunk = self.work.buffer
            if self.plugin:
                chunk = self.plugin.on_response_chunk(chunk)
            try:
                # Call super() for client flush
                teardown = await super().handle_writables(writables)
                if teardown:
                    return True
            except BrokenPipeError:
                logger.warning(     # pragma: no cover
                    'BrokenPipeError when flushing buffer for client',
                )
                return True
            except OSError as exc:
                logger.exception(  # pragma: no cover
                    'OSError when flushing buffer to client',
                    exc_info=exc,
                )
                return True
        return False

    async def handle_readables(self, readables: Readables) -> bool:
        if self.work.connection.fileno() in readables:
            logger.debug('Client is read ready, receiving...')
            self.last_activity = time.time()
            try:
                teardown = await super().handle_readables(readables)
                if teardown:
                    return teardown
            except ssl.SSLWantReadError:    # Try again later
                logger.warning(
                    'SSLWantReadError encountered while reading from client, will retry ...',
                )
                return False
            except socket.error as e:
                if e.errno == errno.ECONNRESET:
                    # Most requests for mobile devices will end up
                    # with client closed connection.  Using `debug`
                    # here to avoid flooding the logs.
                    logger.debug('%r' % e)
                else:
                    logger.warning(
                        'Exception when receiving from %s connection#%d with reason %r' %
                        (self.work.tag, self.work.connection.fileno(), e),
                        exc_info=True,
                    )
                return True
        return False

    ##
    # Internal methods
    ##

    def _initialize_plugin(
            self,
            klass: Type['HttpProtocolHandlerPlugin'],
    ) -> HttpProtocolHandlerPlugin:
        """Initializes passed HTTP protocol handler plugin class."""
        return klass(
            self.uid,
            self.flags,
            self.work,
            self.request,
            self.event_queue,
            self.upstream_conn_pool,
        )

    def _discover_plugin_klass(self, protocol: int) -> Optional[Type['HttpProtocolHandlerPlugin']]:
        """Discovers and return matching HTTP handler plugin matching protocol."""
        if b'HttpProtocolHandlerPlugin' in self.flags.plugins:
            for klass in self.flags.plugins[b'HttpProtocolHandlerPlugin']:
                k: Type['HttpProtocolHandlerPlugin'] = klass
                if protocol in k.protocols():
                    return k
        return None

    def _parse_first_request(self, data: memoryview) -> bool:
        # Parse http request
        try:
            self.request.parse(data)
        except HttpProtocolException as e:  # noqa: WPS329
            self.work.queue(BAD_REQUEST_RESPONSE_PKT)
            raise e
        except Exception as e:
            self.work.queue(BAD_REQUEST_RESPONSE_PKT)
            raise HttpProtocolException(
                'Error when parsing request: %r' % data.tobytes(),
            ) from e
        if not self.request.is_complete:
            return False
        # Bail out if http protocol is unknown
        if self.request.http_handler_protocol == httpProtocols.UNKNOWN:
            self.work.queue(BAD_REQUEST_RESPONSE_PKT)
            return True
        # Discover which HTTP handler plugin is capable of
        # handling the current incoming request
        klass = self._discover_plugin_klass(self.request.http_handler_protocol)
        if klass is None:
            # No matching protocol class found.
            # Return bad request response and
            # close the connection.
            self.work.queue(BAD_REQUEST_RESPONSE_PKT)
            return True
        assert klass is not None
        self.plugin = self._initialize_plugin(klass)
        # Invoke plugin.on_request_complete
        output = self.plugin.on_request_complete()
        if isinstance(output, bool):
            return output
        assert isinstance(output, ssl.SSLSocket)
        logger.debug(
            'Updated client conn to %s', output,
        )
        self.work._conn = output
        return False

    def _connection_inactive_for(self) -> float:
        return time.time() - self.last_activity

    ##
    # run() and _run_once() are here to maintain backward compatibility
    # with threaded mode.  These methods are only called when running
    # in threaded mode.
    ##

    def run(self) -> None:
        """run() method is not used when in --threadless mode.

        This is here just to maintain backward compatibility with threaded mode.
        """
        loop = asyncio.new_event_loop()
        try:
            self.initialize()
            while True:
                # Tear down if client buffer is empty and connection is inactive
                if self.is_inactive():
                    logger.debug(
                        'Client buffer is empty and maximum inactivity has reached '
                        'between client and server connection, tearing down...',
                    )
                    break
                if loop.run_until_complete(self._run_once()):
                    break
        except KeyboardInterrupt:  # pragma: no cover
            pass
        except ssl.SSLError as e:
            logger.exception('ssl.SSLError', exc_info=e)
        except Exception as e:
            logger.exception(
                'Exception while handling connection %r' %
                self.work.connection, exc_info=e,
            )
        finally:
            self.shutdown()
            if self.selector:
                self.selector.close()
            loop.close()

    async def _run_once(self) -> bool:
        events, readables, writables = await self._selected_events()
        try:
            return await self.handle_events(readables, writables)
        finally:
            assert self.selector
            # TODO: Like Threadless we should not unregister
            # work fds repeatedly.
            for fd in events:
                self.selector.unregister(fd)

    # FIXME: Returning events is only necessary because we cannot use async context manager
    # for < Python 3.8.  As a reason, this method is no longer a context manager and caller
    # is responsible for unregistering the descriptors.
    async def _selected_events(self) -> Tuple[SelectableEvents, Readables, Writables]:
        assert self.selector
        events = await self.get_events()
        for fd in events:
            self.selector.register(fd, events[fd])
        ev = self.selector.select(timeout=DEFAULT_SELECTOR_SELECT_TIMEOUT)
        readables = []
        writables = []
        for key, mask in ev:
            if mask & selectors.EVENT_READ:
                readables.append(key.fd)
            if mask & selectors.EVENT_WRITE:
                writables.append(key.fd)
        return (events, readables, writables)

    def _flush(self) -> None:
        assert self.selector
        logger.debug('Flushing pending data')
        try:
            self.selector.register(
                self.work.connection,
                selectors.EVENT_WRITE,
            )
            while self.work.has_buffer():
                logging.debug('Waiting for client read ready')
                ev: List[
                    Tuple[selectors.SelectorKey, int]
                ] = self.selector.select(timeout=DEFAULT_SELECTOR_SELECT_TIMEOUT)
                if len(ev) == 0:
                    continue
                self.work.flush(self.flags.max_sendbuf_size)
        except BrokenPipeError:
            pass
        finally:
            self.selector.unregister(self.work.connection)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
"""
import gzip
from typing import Dict, List, Type, Tuple, TypeVar, Optional

from ..url import Url
from .chunk import ChunkParser, chunkParserStates
from .types import httpParserTypes, httpParserStates
from ..methods import httpMethods
from .protocol import ProxyProtocol
from ..exception import HttpProtocolException
from ..protocols import httpProtocols
from ...common.flag import flags
from ...common.utils import (
    text_, bytes_, build_http_request, build_http_response,
)
from ...common.constants import (
    CRLF, COLON, SLASH, HTTP_1_0, HTTP_1_1, WHITESPACE, DEFAULT_HTTP_PORT,
    DEFAULT_DISABLE_HEADERS, DEFAULT_ENABLE_PROXY_PROTOCOL,
)


flags.add_argument(
    '--enable-proxy-protocol',
    action='store_true',
    default=DEFAULT_ENABLE_PROXY_PROTOCOL,
    help='Default: ' + str(DEFAULT_ENABLE_PROXY_PROTOCOL) + '.  ' +
    'If used, will enable proxy protocol.  ' +
    'Only version 1 is currently supported.',
)


T = TypeVar('T', bound='HttpParser')


class HttpParser:
    """HTTP request/response parser.

    TODO: Make me zero-copy by using :class:`memoryview`.
    Currently due to chunk/buffer handling we
    are not able to utilize :class:`memoryview`
    efficiently.

    For this to happen we must store ``buffer``
    as ``List[memoryview]`` instead of raw bytes and
    update parser to work accordingly.
    """

    def __init__(
            self, parser_type: int,
            enable_proxy_protocol: int = DEFAULT_ENABLE_PROXY_PROTOCOL,
    ) -> None:
        self.state: int = httpParserStates.INITIALIZED
        self.type: int = parser_type
        self.protocol: Optional[ProxyProtocol] = None
        if enable_proxy_protocol:
            assert self.type == httpParserTypes.REQUEST_PARSER
            self.protocol = ProxyProtocol()
        # Request attributes
        self.host: Optional[bytes] = None
        self.port: Optional[int] = None
        self.path: Optional[bytes] = None
        self.method: Optional[bytes] = None
        # Response attributes
        self.code: Optional[bytes] = None
        self.reason: Optional[bytes] = None
        self.version: Optional[bytes] = None
        # Total size of raw bytes passed for parsing
        self.total_size: int = 0
        # Buffer to hold unprocessed bytes
        self.buffer: Optional[memoryview] = None
        # Internal headers data structure:
        # - Keys are lower case header names.
        # - Values are 2-tuple containing original
        #   header and it's value as received.
        self.headers: Optional[Dict[bytes, Tuple[bytes, bytes]]] = None
        self.body: Optional[bytes] = None
        self.chunk: Optional[ChunkParser] = None
        # Internal request line as a url structure
        self._url: Optional[Url] = None
        # Deduced states from the packet
        self._is_chunked_encoded: bool = False
        self._content_expected: bool = False
        self._is_https_tunnel: bool = False

    @classmethod
    def request(
            cls: Type[T],
            raw: bytes,
            enable_proxy_protocol: int = DEFAULT_ENABLE_PROXY_PROTOCOL,
    ) -> T:
        parser = cls(
            httpParserTypes.REQUEST_PARSER,
            enable_proxy_protocol=enable_proxy_protocol,
        )
        parser.parse(memoryview(raw))
        return parser

    @classmethod
    def response(cls: Type[T], raw: bytes) -> T:
        parser = cls(httpParserTypes.RESPONSE_PARSER)
        parser.parse(memoryview(raw))
        return parser

    def header(self, key: bytes) -> bytes:
        """Convenient method to return original header value from internal data structure."""
        if self.headers is None or key.lower() not in self.headers:
            raise KeyError('%s not found in headers' % text_(key))
        return self.headers[key.lower()][1]

    def has_header(self, key: bytes) -> bool:
        """Returns true if header key was found in payload."""
        if self.headers is None:
            return False
        return key.lower() in self.headers

    def add_header(self, key: bytes, value: bytes) -> bytes:
        """Add/Update a header to internal data structure.

        Returns key with which passed (key, value) tuple is available."""
        if self.headers is None:
            self.headers = {}
        k = key.lower()
        # k = key
        self.headers[k] = (key, value)
        return k

    def add_headers(self, headers: List[Tuple[bytes, bytes]]) -> None:
        """Add/Update multiple headers to internal data structure"""
        for (key, value) in headers:
            self.add_header(key, value)

    def del_header(self, header: bytes) -> None:
        """Delete a header from internal data structure."""
        if self.headers and header.lower() in self.headers:
            del self.headers[header.lower()]

    def del_headers(self, headers: List[bytes]) -> None:
        """Delete headers from internal data structure."""
        for key in headers:
            self.del_header(key.lower())

    def set_url(self, url: bytes, allowed_url_schemes: Optional[List[bytes]] = None) -> None:
        """Given a request line, parses it and sets line attributes a.k.a. host, port, path."""
        self._url = Url.from_bytes(
            url, allowed_url_schemes=allowed_url_schemes,
        )
        self._set_line_attributes()

    def update_body(self, body: bytes, content_type: bytes) -> None:
        """This method must be used to update body after HTTP packet has been parsed.

        Along with updating the body, this method also respects original
        request content encoding, transfer encoding settings."""
        # If outgoing request encoding is gzip
        # also compress the body
        if self.has_header(b'content-encoding'):
            if self.header(b'content-encoding') == b'gzip':
                body = gzip.compress(body)
            else:
                # We only work with gzip, for any other encoding
                # type, remove the original header
                self.del_header(b'content-encoding')
        # If the request is of type chunked encoding
        # add post data as chunk
        if self.is_chunked_encoded:
            body = ChunkParser.to_chunks(body)
            self.del_header(b'content-length')
        else:
            self.add_header(
                b'Content-Length',
                bytes_(len(body)),
            )
        self.body = body
        self.add_header(b'Content-Type', content_type)

    @property
    def http_handler_protocol(self) -> int:
        """Returns `HttpProtocols` that this request belongs to."""
        if self.version in (HTTP_1_1, HTTP_1_0) and self._url is not None:
            if self.host is not None:
                return httpProtocols.HTTP_PROXY
            if self._url.hostname is None:
                return httpProtocols.WEB_SERVER
        return httpProtocols.UNKNOWN

    @property
    def is_complete(self) -> bool:
        return self.state == httpParserStates.COMPLETE

    @property
    def is_http_1_1_keep_alive(self) -> bool:
        """Returns true for HTTP/1.1 keep-alive connections."""
        return self.version == HTTP_1_1 and \
            (
                not self.has_header(b'Connection') or
                self.header(b'Connection').lower() == b'keep-alive'
            )

    @property
    def is_connection_upgrade(self) -> bool:
        """Returns true for websocket upgrade requests."""
        return self.version == HTTP_1_1 and \
            self.has_header(b'Connection') and \
            self.has_header(b'Upgrade')

    @property
    def is_websocket_upgrade(self) -> bool:
        return self.is_connection_upgrade and self.header(b'upgrade').lower() in (
            b'websocket',
            b'derp',
        )

    @property
    def is_https_tunnel(self) -> bool:
        """Returns true for HTTPS CONNECT tunnel request."""
        return self._is_https_tunnel

    @property
    def is_chunked_encoded(self) -> bool:
        """Returns true if transfer-encoding chunked is used."""
        return self._is_chunked_encoded

    @property
    def content_expected(self) -> bool:
        """Returns true if content-length is present and not 0."""
        return self._content_expected

    @property
    def body_expected(self) -> bool:
        """Returns true if content or chunked response is expected."""
        return self._content_expected or self._is_chunked_encoded

    def parse(
            self,
            raw: memoryview,
            allowed_url_schemes: Optional[List[bytes]] = None,
    ) -> None:
        """Parses HTTP request out of raw bytes.

        Check for `HttpParser.state` after `parse` has successfully returned."""
        size = len(raw)
        self.total_size += size
        if self.buffer:
            # TODO(abhinavsingh): Instead of tobytes our parser
            # must be capable of working with arrays of memoryview
            raw = memoryview(self.buffer.tobytes() + raw.tobytes())
        self.buffer, more = None, size > 0
        while more and self.state != httpParserStates.COMPLETE:
            # gte with HEADERS_COMPLETE also encapsulated RCVING_BODY state
            if self.state >= httpParserStates.HEADERS_COMPLETE:
                more, raw = self._process_body(raw)
            elif self.state == httpParserStates.INITIALIZED:
                more, raw = self._process_line(
                    raw,
                    allowed_url_schemes=allowed_url_schemes,
                )
            else:
                more, raw = self._process_headers(raw)
            # When server sends a response line without any header or body e.g.
            # HTTP/1.1 200 Connection established\r\n\r\n
            if self.type == httpParserTypes.RESPONSE_PARSER and \
                    self.state == httpParserStates.LINE_RCVD and \
                    raw == CRLF:
                self.state = httpParserStates.COMPLETE
            # Mark request as complete if headers received and no incoming
            # body indication received.
            elif self.state == httpParserStates.HEADERS_COMPLETE and \
                    not (self._content_expected or self._is_chunked_encoded) and \
                    raw == b'':
                self.state = httpParserStates.COMPLETE
        self.buffer = None if raw == b'' else raw

    def build(
        self,
        disable_headers: Optional[List[bytes]] = None,
        for_proxy: bool = False,
        host: Optional[bytes] = None,
    ) -> bytes:
        """Rebuild the request object."""
        assert self.method and self.version and self.type == httpParserTypes.REQUEST_PARSER
        if disable_headers is None:
            disable_headers = DEFAULT_DISABLE_HEADERS
        body: Optional[bytes] = self._get_body_or_chunks()
        path = self.path or b'/'
        if for_proxy:
            assert self.host and self.port and self._url
            path = (
                b'http' if not self._url.scheme else self._url.scheme +
                COLON + SLASH + SLASH +
                self.host +
                COLON +
                str(self.port).encode() +
                path
            ) if not self._is_https_tunnel else (self.host + COLON + str(self.port).encode())
        return build_http_request(
            self.method,
            path,
            self.version,
            headers=(
                {}
                if not self.headers
                else {
                    self.headers[k][0]: (
                        self.headers[k][1]
                        if host is None or self.headers[k][0].lower() != b'host'
                        else host
                    )
                    for k in self.headers
                    if k.lower() not in disable_headers
                }
            ),
            body=body,
            no_ua=True,
        )

    def build_response(self) -> bytes:
        """Rebuild the response object."""
        assert self.code and self.version and self.type == httpParserTypes.RESPONSE_PARSER
        return build_http_response(
            status_code=int(self.code),
            protocol_version=self.version,
            reason=self.reason,
            headers={} if not self.headers else {
                self.headers[k][0]: self.headers[k][1] for k in self.headers
            },
            body=self._get_body_or_chunks(),
        )

    def _process_body(self, raw: memoryview) -> Tuple[bool, memoryview]:
        # Ref: http://www.ietf.org/rfc/rfc2616.txt
        # 3.If a Content-Length header field (section 14.13) is present, its
        #   decimal value in OCTETs represents both the entity-length and the
        #   transfer-length. The Content-Length header field MUST NOT be sent
        #   if these two lengths are different (i.e., if a Transfer-Encoding
        #   header field is present). If a message is received with both a
        #   Transfer-Encoding header field and a Content-Length header field,
        #   the latter MUST be ignored.
        #
        # TL;DR -- Give transfer-encoding header preference over content-length.
        if self._is_chunked_encoded:
            if not self.chunk:
                self.chunk = ChunkParser()
            raw = self.chunk.parse(raw)
            if self.chunk.state == chunkParserStates.COMPLETE:
                self.body = self.chunk.body
                self.state = httpParserStates.COMPLETE
            more = False
            return more, raw
        if self._content_expected:
            self.state = httpParserStates.RCVING_BODY
            if self.body is None:
                self.body = b''
            total_size = int(self.header(b'content-length'))
            received_size = len(self.body)
            self.body += raw[:total_size - received_size]
            if self.body and \
                    len(self.body) == int(self.header(b'content-length')):
                self.state = httpParserStates.COMPLETE
            return len(raw) > 0, raw[total_size - received_size:]
        # Received a packet without content-length header
        # and no transfer-encoding specified.
        #
        # This can happen for both HTTP/1.0 and HTTP/1.1 scenarios.
        # Currently, we consume the remaining buffer as body.
        #
        # Ref https://github.com/abhinavsingh/proxy.py/issues/398
        #
        # See TestHttpParser.test_issue_398 scenario
        self.state = httpParserStates.RCVING_BODY
        self.body = bytes(raw)
        return False, memoryview(b'')

    def _process_headers(self, raw: memoryview) -> Tuple[bool, memoryview]:
        """Returns False when no CRLF could be found in received bytes.

        TODO: We should not return until parser reaches headers complete
        state or when there is no more data left to parse.

        TODO: For protection against Slowloris attack, we must parse the
        request line and headers only after receiving end of header marker.
        This will also help make the parser even more stateless.
        """
        while True:
            parts = raw.tobytes().split(CRLF, 1)
            if len(parts) == 1:
                return False, raw
            line, raw = parts[0], memoryview(parts[1])
            if self.state in (httpParserStates.LINE_RCVD, httpParserStates.RCVING_HEADERS):
                if line == b'' or line.strip() == b'':  # Blank line received.
                    self.state = httpParserStates.HEADERS_COMPLETE
                else:
                    self.state = httpParserStates.RCVING_HEADERS
                    self._process_header(line)
            # If raw length is now zero, bail out
            # If we have received all headers, bail out
            if raw == b'' or self.state == httpParserStates.HEADERS_COMPLETE:
                break
        return len(raw) > 0, raw

    def _process_line(
            self,
            raw: memoryview,
            allowed_url_schemes: Optional[List[bytes]] = None,
    ) -> Tuple[bool, memoryview]:
        while True:
            parts = raw.tobytes().split(CRLF, 1)
            if len(parts) == 1:
                return False, raw
            line, raw = parts[0], memoryview(parts[1])
            if self.type == httpParserTypes.REQUEST_PARSER:
                if self.protocol is not None and self.protocol.version is None:
                    # We expect to receive entire proxy protocol v1 line
                    # in one network read and don't expect partial packets
                    self.protocol.parse(line)
                    continue
                # Ref: https://datatracker.ietf.org/doc/html/rfc2616#section-5.1
                parts = line.split(WHITESPACE, 2)
                if len(parts) == 3:
                    self.method = parts[0]
                    if self.method == httpMethods.CONNECT:
                        self._is_https_tunnel = True
                    self.set_url(
                        parts[1], allowed_url_schemes=allowed_url_schemes,
                    )
                    self.version = parts[2]
                    self.state = httpParserStates.LINE_RCVD
                    break
                # To avoid a possible attack vector, we raise exception
                # if parser receives an invalid request line.
                raise HttpProtocolException(
                    'Invalid request line %r' % raw.tobytes(),
                )
            parts = line.split(WHITESPACE, 2)
            self.version = parts[0]
            self.code = parts[1]
            # Our own WebServerPlugin example currently doesn't send any reason
            if len(parts) == 3:
                self.reason = parts[2]
            self.state = httpParserStates.LINE_RCVD
            break
        return len(raw) > 0, raw

    def _process_header(self, raw: bytes) -> None:
        parts = raw.split(COLON, 1)
        key, value = (
            parts[0].strip(),
            b'' if len(parts) == 1 else parts[1].strip(),
        )
        k = self.add_header(key, value)
        # b'content-length' in self.headers and int(self.header(b'content-length')) > 0
        if k == b'content-length' and int(value) > 0:
            self._content_expected = True
        # return b'transfer-encoding' in self.headers and \
        #   self.headers[b'transfer-encoding'][1].lower() == b'chunked'
        elif k == b'transfer-encoding' and value.lower() == b'chunked':
            self._is_chunked_encoded = True

    def _get_body_or_chunks(self) -> Optional[bytes]:
        return ChunkParser.to_chunks(self.body) \
            if self.body and self._is_chunked_encoded else \
            self.body

    def _set_line_attributes(self) -> None:
        if self.type == httpParserTypes.REQUEST_PARSER:
            assert self._url
            if self._is_https_tunnel:
                self.host = self._url.hostname
                self.port = 443 if self._url.port is None else self._url.port
            else:
                self.host, self.port = self._url.hostname, self._url.port \
                    if self._url.port else DEFAULT_HTTP_PORT
            self.path = self._url.remainder



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import logging
from typing import BinaryIO, Optional

from .base import CacheStore
from ....common.flag import flags
from ....http.parser import HttpParser
from ....common.utils import text_
from ....common.constants import DEFAULT_CACHE_DIRECTORY_PATH


logger = logging.getLogger(__name__)


flags.add_argument(
    '--cache-dir',
    type=str,
    default=DEFAULT_CACHE_DIRECTORY_PATH,
    help='Default: ' + DEFAULT_CACHE_DIRECTORY_PATH + '.  ' +
    'Flag only applicable when cache plugin is used with on-disk storage.',
)


class OnDiskCacheStore(CacheStore):

    def __init__(self, uid: str, cache_dir: str, cache_requests: bool) -> None:
        super().__init__(uid)
        self.cache_dir = cache_dir
        self.cache_requests = cache_requests
        self.cache_file_name: Optional[str] = None
        self.cache_file_path: Optional[str] = None
        self.cache_file: Optional[BinaryIO] = None

    def open(self, request: HttpParser) -> None:
        self.cache_file_name = '%s-%s' % (text_(request.host), self.uid)
        self.cache_file_path = os.path.join(
            self.cache_dir,
            '%s.txt' % self.cache_file_name,
        )
        self.cache_file = open(self.cache_file_path, "wb")

    def cache_request(self, request: HttpParser) -> Optional[HttpParser]:
        if self.cache_file and self.cache_requests:
            self.cache_file.write(request.build())
        return request

    def cache_response_chunk(self, chunk: memoryview) -> memoryview:
        if self.cache_file:
            self.cache_file.write(chunk.tobytes())
        return chunk

    def close(self) -> None:
        if self.cache_file:
            self.cache_file.flush()
            self.cache_file.close()
            self.cache_file = None
            logger.debug('Cached response at %s', self.cache_file_path)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import queue
import asyncio
import contextlib
from abc import abstractmethod
from typing import Any, Optional

from .threadless import Threadless
from ...common.backports import NonBlockingQueue


class BaseLocalExecutor(Threadless[NonBlockingQueue]):
    """A threadless executor implementation which uses a queue to receive new work."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._loop: Optional[asyncio.AbstractEventLoop] = None

    @property
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        if self._loop is None:
            self._loop = asyncio.get_event_loop_policy().new_event_loop()
        return self._loop

    def work_queue_fileno(self) -> Optional[int]:
        return None

    def receive_from_work_queue(self) -> bool:
        with contextlib.suppress(queue.Empty):
            work = self.work_queue.get()
            if isinstance(work, bool) and work is False:
                return True
            self.work(work)
        return False

    @abstractmethod
    def work(self, *args: Any) -> None:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import Optional
from urllib import parse as urlparse

from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser


class RedirectToCustomServerPlugin(HttpProxyBasePlugin):
    """Modifies client request to redirect all incoming requests to a fixed server address."""

    UPSTREAM_SERVER = b'http://localhost:8899/'

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        # Redirect all non-https requests to inbuilt WebServer.
        if not request.is_https_tunnel:
            request.set_url(self.UPSTREAM_SERVER)
            # Update Host header too, otherwise upstream can reject our request
            if request.has_header(b'Host'):
                request.del_header(b'Host')
            request.add_header(
                b'Host', urlparse.urlsplit(
                    self.UPSTREAM_SERVER,
                ).netloc,
            )
        return request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import logging
from typing import List, Tuple

from ..http.parser import HttpParser
from ..http.server import HttpWebServerBasePlugin, httpProtocolTypes
from ..http.responses import permanentRedirectResponse


logger = logging.getLogger(__name__)


class ProxyDashboard(HttpWebServerBasePlugin):
    """Proxy Dashboard."""

    # Redirects to /dashboard/
    REDIRECT_ROUTES = [
        (httpProtocolTypes.HTTP, r'/dashboard$'),
        (httpProtocolTypes.HTTPS, r'/dashboard$'),
        (httpProtocolTypes.HTTP, r'/dashboard/proxy.html$'),
        (httpProtocolTypes.HTTPS, r'/dashboard/proxy.html$'),
    ]

    # Index html route
    INDEX_ROUTES = [
        (httpProtocolTypes.HTTP, r'/dashboard/$'),
        (httpProtocolTypes.HTTPS, r'/dashboard/$'),
    ]

    def routes(self) -> List[Tuple[int, str]]:
        return ProxyDashboard.REDIRECT_ROUTES + \
            ProxyDashboard.INDEX_ROUTES

    def handle_request(self, request: HttpParser) -> None:
        if request.path == b'/dashboard/':
            self.client.queue(
                self.serve_static_file(
                    os.path.join(
                        self.flags.static_server_dir,
                        'dashboard', 'proxy.html',
                    ),
                    self.flags.min_compression_length,
                ),
            )
        elif request.path in (
                b'/dashboard',
                b'/dashboard/proxy.html',
        ):
            self.client.queue(permanentRedirectResponse(b'/dashboard/'))



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import logging
import argparse
import multiprocessing
from typing import TYPE_CHECKING, Any, List, Type, TypeVar, Optional
from multiprocessing import connection

from ...common.flag import flags
from ...common.constants import DEFAULT_THREADLESS, DEFAULT_NUM_WORKERS


if TYPE_CHECKING:   # pragma: no cover
    from ..event import EventQueue
    from .threadless import Threadless

T = TypeVar('T', bound='Threadless[Any]')

logger = logging.getLogger(__name__)


flags.add_argument(
    '--threadless',
    action='store_true',
    default=DEFAULT_THREADLESS,
    help='Default: ' + ('True' if DEFAULT_THREADLESS else 'False') + '.  ' +
    'Enabled by default on Python 3.8+ (mac, linux).  ' +
    'When disabled a new thread is spawned '
    'to handle each client connection.',
)

flags.add_argument(
    '--threaded',
    action='store_true',
    default=not DEFAULT_THREADLESS,
    help='Default: ' + ('True' if not DEFAULT_THREADLESS else 'False') + '.  ' +
    'Disabled by default on Python < 3.8 and windows.  ' +
    'When enabled a new thread is spawned '
    'to handle each client connection.',
)

flags.add_argument(
    '--num-workers',
    type=int,
    default=DEFAULT_NUM_WORKERS,
    help='Defaults to number of CPU cores.',
)


class ThreadlessPool:
    """Manages lifecycle of threadless pool and delegates work to them
    using a round-robin strategy.

    Example usage::

        with ThreadlessPool(flags=...) as pool:
            while True:
                time.sleep(1)

    If necessary, start multiple threadless pool with different
    work classes.
    """

    def __init__(
        self,
        flags: argparse.Namespace,
        executor_klass: Type['T'],
        event_queue: Optional['EventQueue'] = None,
    ) -> None:
        self.flags = flags
        self.event_queue = event_queue
        # Threadless worker communication states
        self.work_queues: List[connection.Connection] = []
        self.work_pids: List[int] = []
        self.work_locks: List['multiprocessing.synchronize.Lock'] = []
        # List of threadless workers
        self._executor_klass = executor_klass
        # FIXME: Instead of Any type must be the executor klass
        self._workers: List[Any] = []
        self._processes: List[multiprocessing.Process] = []

    def __enter__(self) -> 'ThreadlessPool':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def setup(self) -> None:
        """Setup threadless processes."""
        if self.flags.threadless:
            for index in range(self.flags.num_workers):
                self._start_worker(index)
            logger.info(
                'Started {0} threadless workers'.format(
                    self.flags.num_workers,
                ),
            )

    def shutdown(self) -> None:
        """Shutdown threadless processes."""
        if self.flags.threadless:
            self._shutdown_workers()
            logger.info(
                'Stopped {0} threadless workers'.format(
                    self.flags.num_workers,
                ),
            )

    def _start_worker(self, index: int) -> None:
        """Starts a threadless worker."""
        self.work_locks.append(multiprocessing.Lock())
        pipe = multiprocessing.Pipe()
        self.work_queues.append(pipe[0])
        w = self._executor_klass(
            iid=str(index),
            work_queue=pipe[1],
            flags=self.flags,
            event_queue=self.event_queue,
        )
        self._workers.append(w)
        p = multiprocessing.Process(target=w.run)
        # p.daemon = True
        self._processes.append(p)
        p.start()
        assert p.pid
        self.work_pids.append(p.pid)
        logger.debug('Started threadless#%d process#%d', index, p.pid)

    def _shutdown_workers(self) -> None:
        """Pop a running threadless worker and clean it up."""
        for index in range(self.flags.num_workers):
            self._workers[index].running.set()
        for _ in range(self.flags.num_workers):
            pid = self.work_pids[-1]
            self._processes.pop().join()
            self._workers.pop()
            self.work_pids.pop()
            self.work_queues.pop().close()
            logger.debug('Stopped threadless process#%d', pid)
        self.work_locks = []



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import socket
import logging
import ipaddress
from typing import Any, Union, Optional

from .base import BaseListener
from ...common.flag import flags
from ...common.constants import (
    DEFAULT_PORT, DEFAULT_PORT_FILE, DEFAULT_IPV4_HOSTNAME,
)


flags.add_argument(
    '--hostname',
    type=str,
    default=str(DEFAULT_IPV4_HOSTNAME),
    help='Default: 127.0.0.1. Server IP address.',
)

flags.add_argument(
    '--hostnames',
    action='append',
    nargs='+',
    type=str,
    default=None,
    help='Default: None.  Additional IP addresses to listen on.',
)

flags.add_argument(
    '--port',
    type=int,
    default=DEFAULT_PORT,
    help='Default: 8899.  Server port.  To listen on more ports, pass them using --ports flag.',
)

flags.add_argument(
    '--ports',
    action='append',
    nargs='+',
    type=int,
    default=None,
    help='Default: None.  Additional ports to listen on.',
)

flags.add_argument(
    '--port-file',
    type=str,
    default=DEFAULT_PORT_FILE,
    help='Default: None. Save server port numbers. Useful when using --port=0 ephemeral mode.',
)

logger = logging.getLogger(__name__)


class TcpSocketListener(BaseListener):
    """Tcp listener."""

    def __init__(
        self,
        hostname: Union[ipaddress.IPv4Address, ipaddress.IPv6Address],
        port: int,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.hostname = hostname
        self.port = port
        # Set after binding to a port.
        #
        # Stored here separately for ephemeral port discovery.
        self._port: Optional[int] = None
        super().__init__(*args, **kwargs)

    def listen(self) -> socket.socket:
        sock = socket.socket(
            socket.AF_INET6 if self.hostname.version == 6 else socket.AF_INET,
            socket.SOCK_STREAM,
        )
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        # s.setsockopt(socket.SOL_TCP, socket.TCP_FASTOPEN, 5)
        sock.bind((str(self.hostname), self.port))
        sock.listen(self.flags.backlog)
        sock.setblocking(False)
        self._port = sock.getsockname()[1]
        logger.debug(
            'Listening on %s:%s' % (self.hostname, self._port),
        )
        return sock



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import sys
import base64
import argparse
import ipaddress
import itertools
import collections
import multiprocessing
from typing import Any, List, Optional, cast

from .types import IpAddress
from .utils import bytes_, is_py2, is_threadless, set_open_file_limit
from .logger import Logger
from .plugins import Plugins
from .version import __version__
from .constants import (
    COMMA, PLUGIN_METRICS, PLUGIN_PAC_FILE, PLUGIN_DASHBOARD,
    PLUGIN_HTTP_PROXY, PLUGIN_PROXY_AUTH, PLUGIN_WEB_SERVER,
    DEFAULT_NUM_WORKERS, PLUGIN_REVERSE_PROXY, DEFAULT_NUM_ACCEPTORS,
    PLUGIN_INSPECT_TRAFFIC, DEFAULT_DISABLE_HEADERS, PY2_DEPRECATION_MESSAGE,
    DEFAULT_DEVTOOLS_WS_PATH, PLUGIN_DEVTOOLS_PROTOCOL,
    PLUGIN_WEBSOCKET_TRANSPORT, DEFAULT_DATA_DIRECTORY_PATH,
    DEFAULT_MIN_COMPRESSION_LENGTH,
)


__homepage__ = 'https://github.com/abhinavsingh/proxy.py'


# TODO: Currently `initialize` staticmethod contains knowledge
# about several common flags defined by proxy.py core.

# This logic must be decoupled.  flags.add_argument must
# also provide a callback to resolve the final flag value
# based upon availability in input_args, **opts and
# default values.

# Supporting such a framework is complex but achievable.
# One problem is that resolution of certain flags
# can depend upon availability of other flags.

# This will lead us into dependency graph modeling domain.
class FlagParser:
    """Wrapper around argparse module.

    Import `flag.flags` and use `add_argument` API
    to define custom flags within respective Python files.

    Best Practice::

       1. Define flags at the top of your class files.
       2. DO NOT add flags within your class `__init__` method OR
          within class methods.  It MAY result into runtime exception,
          especially if your class is initialized multiple times or if
          class method registering the flag gets invoked multiple times.

    """

    def __init__(self) -> None:
        self.args: Optional[argparse.Namespace] = None
        self.actions: List[str] = []
        self.parser = argparse.ArgumentParser(
            description='proxy.py v%s' % __version__,
            epilog='Proxy.py not working? Report at: %s/issues/new' % __homepage__,
        )

    def add_argument(self, *args: Any, **kwargs: Any) -> argparse.Action:
        """Register a flag."""
        action = self.parser.add_argument(*args, **kwargs)
        self.actions.append(action.dest)
        return action

    def parse_args(
            self, input_args: Optional[List[str]],
    ) -> argparse.Namespace:
        """Parse flags from input arguments."""
        self.args = self.parser.parse_args(input_args)
        return self.args

    @staticmethod
    def initialize(
        input_args: Optional[List[str]] = None,
        **opts: Any,
    ) -> argparse.Namespace:
        if input_args is None:
            input_args = []

        if is_py2():
            print(PY2_DEPRECATION_MESSAGE)
            sys.exit(1)

        # Discover flags from requested plugin.
        # This will also surface external plugin flags
        # under --help.
        Plugins.discover(input_args)

        # Parse flags
        args = flags.parse_args(input_args)

        # Print version and exit
        if args.version:
            print(__version__)
            sys.exit(0)

        # proxy.py currently cannot serve over HTTPS and also perform TLS interception
        # at the same time.  Check if user is trying to enable both feature
        # at the same time.
        #
        # TODO: Use parser.add_mutually_exclusive_group()
        # and remove this logic from here.
        if (args.cert_file and args.key_file) and \
                (args.ca_key_file and args.ca_cert_file and args.ca_signing_key_file):
            print(
                'You can either enable end-to-end encryption OR TLS interception,'
                'not both together.',
            )
            sys.exit(1)

        # Setup logging module
        Logger.setup(args.log_file, args.log_level, args.log_format)

        # Setup limits
        set_open_file_limit(args.open_file_limit)

        # Load work_klass
        work_klass = opts.get('work_klass', args.work_klass)
        work_klass = Plugins.importer(bytes_(work_klass))[0] \
            if isinstance(work_klass, str) \
            else work_klass

        # TODO: Plugin flag initialization logic must be moved within plugins.
        #
        # Generate auth_code required for basic authentication if enabled
        auth_code = None
        basic_auth = opts.get('basic_auth', getattr(args, 'basic_auth', None))
        # Destroy passed credentials via flags or options
        args.basic_auth = None
        if 'basic_auth' in opts:
            del opts['basic_auth']

        # Resolve auth module.
        auth_plugins = []
        auth_plugin = opts.get('auth_plugin', args.auth_plugin)
        if basic_auth:
            auth_code = base64.b64encode(bytes_(basic_auth))
        if basic_auth or auth_plugin != PLUGIN_PROXY_AUTH:
            # No basic auth provided
            # Here auth_plugin is set to default plugin
            # We want to avoid loading the auth plugin (w/o basic auth)
            # unless user overrides the default auth plugin.
            auth_plugins.append(auth_plugin)

        # --enable flags must be parsed before loading plugins
        # otherwise we will miss the plugins passed via constructor
        args.enable_web_server = cast(
            bool,
            opts.get(
                'enable_web_server',
                args.enable_web_server,
            ),
        )
        args.enable_static_server = cast(
            bool,
            opts.get(
                'enable_static_server',
                args.enable_static_server,
            ),
        )
        args.enable_events = cast(
            bool,
            opts.get(
                'enable_events',
                args.enable_events,
            ),
        )
        args.enable_metrics = cast(
            bool,
            opts.get(
                'enable_metrics',
                args.enable_metrics,
            ),
        )

        # Load default plugins along with user provided --plugins
        default_plugins = [
            bytes_(p)
            for p in FlagParser.get_default_plugins(args)
        ]
        requested_plugins = Plugins.resolve_plugin_flag(
            args.plugins, opts.get('plugins'),
        )
        plugins = Plugins.load(
            default_plugins + auth_plugins + requested_plugins,
        )

        if bytes_(PLUGIN_METRICS) in default_plugins:
            args.metrics_lock = multiprocessing.Lock()

        # https://github.com/python/mypy/issues/5865
        #
        # def option(t: object, key: str, default: Any) -> Any:
        #     return cast(t, opts.get(key, default))
        args.work_klass = work_klass
        args.plugins = plugins
        args.auth_code = cast(
            Optional[bytes],
            opts.get(
                'auth_code',
                auth_code,
            ),
        )
        args.server_recvbuf_size = cast(
            int,
            opts.get(
                'server_recvbuf_size',
                args.server_recvbuf_size,
            ),
        )
        args.client_recvbuf_size = cast(
            int,
            opts.get(
                'client_recvbuf_size',
                args.client_recvbuf_size,
            ),
        )
        args.pac_file = cast(
            Optional[str], opts.get(
                'pac_file', bytes_(
                    args.pac_file,
                ),
            ),
        )
        args.pac_file_url_path = cast(
            Optional[bytes], opts.get(
                'pac_file_url_path', bytes_(
                    args.pac_file_url_path,
                ),
            ),
        )
        disabled_headers = cast(
            Optional[List[bytes]], opts.get(
                'disable_headers', [
                    header.lower()
                    for header in bytes_(args.disable_headers).split(COMMA)
                    if header.strip() != b''
                ],
            ),
        )
        args.disable_headers = disabled_headers if disabled_headers is not None else DEFAULT_DISABLE_HEADERS

        args.certfile = cast(
            Optional[str], opts.get(
                'cert_file', args.cert_file,
            ),
        )
        args.keyfile = cast(Optional[str], opts.get('key_file', args.key_file))

        args.ca_key_file = cast(
            Optional[str], opts.get(
                'ca_key_file', args.ca_key_file,
            ),
        )
        args.ca_cert_file = cast(
            Optional[str], opts.get(
                'ca_cert_file', args.ca_cert_file,
            ),
        )
        args.ca_signing_key_file = cast(
            Optional[str],
            opts.get(
                'ca_signing_key_file',
                args.ca_signing_key_file,
            ),
        )
        args.ca_file = cast(
            Optional[str],
            opts.get(
                'ca_file',
                args.ca_file,
            ),
        )
        args.openssl = cast(
            Optional[str],
            opts.get(
                'openssl',
                args.openssl,
            ),
        )

        args.hostname = cast(
            IpAddress,
            opts.get('hostname', ipaddress.ip_address(args.hostname)),
        )
        hostnames: List[List[str]] = opts.get('hostnames', args.hostnames)
        args.hostnames = [
            ipaddress.ip_address(hostname) for hostname in list(
                itertools.chain.from_iterable([] if hostnames is None else hostnames),
            )
        ]
        args.unix_socket_path = opts.get(
            'unix_socket_path', args.unix_socket_path,
        )
        args.port = cast(int, opts.get('port', args.port))
        ports: List[List[int]] = opts.get('ports', args.ports)
        args.ports = [
            int(port) for port in list(
                itertools.chain.from_iterable([] if ports is None else ports),
            )
        ]
        args.backlog = cast(int, opts.get('backlog', args.backlog))
        num_workers = opts.get('num_workers', args.num_workers)
        args.num_workers = cast(
            int, num_workers if num_workers > 0 else multiprocessing.cpu_count(),
        )
        num_acceptors = opts.get('num_acceptors', args.num_acceptors)
        # See https://github.com/abhinavsingh/proxy.py/pull/714 description
        # to understand rationale behind the following logic.
        #
        # Num workers flag or option was found. We will use
        # the same value for num_acceptors when num acceptors flag
        # is absent.
        if num_workers != DEFAULT_NUM_WORKERS and num_acceptors == DEFAULT_NUM_ACCEPTORS:
            args.num_acceptors = args.num_workers
        else:
            args.num_acceptors = cast(
                int, num_acceptors if num_acceptors > 0 else multiprocessing.cpu_count(),
            )

        args.static_server_dir = cast(
            str,
            opts.get(
                'static_server_dir',
                args.static_server_dir,
            ),
        )
        args.min_compression_length = cast(
            bool,
            opts.get(
                'min_compression_length',
                getattr(
                    args, 'min_compression_length',
                    DEFAULT_MIN_COMPRESSION_LENGTH,
                ),
            ),
        )
        args.devtools_ws_path = cast(
            bytes,
            opts.get(
                'devtools_ws_path',
                getattr(args, 'devtools_ws_path', DEFAULT_DEVTOOLS_WS_PATH),
            ),
        )
        args.timeout = cast(int, opts.get('timeout', args.timeout))
        args.local_executor = cast(
            int,
            opts.get(
                'local_executor',
                args.local_executor,
            ),
        )
        args.threaded = cast(bool, opts.get('threaded', args.threaded))
        # Pre-evaluate threadless values based upon environment and config
        #
        # --threadless is now default mode of execution
        # but we still have exceptions based upon OS config.
        # Make sure executors are not started if is_threadless
        # evaluates to False.
        args.threadless = cast(bool, opts.get('threadless', args.threadless))
        args.threadless = is_threadless(args.threadless, args.threaded)

        args.pid_file = cast(
            Optional[str], opts.get(
                'pid_file',
                args.pid_file,
            ),
        )

        args.port_file = cast(
            Optional[str], opts.get(
                'port_file',
                args.port_file,
            ),
        )

        args.proxy_py_data_dir = cast(
            str,
            opts.get(
                'data_dir',
                args.data_dir or DEFAULT_DATA_DIRECTORY_PATH,
            ),
        )
        os.makedirs(args.proxy_py_data_dir, exist_ok=True)

        ca_cert_dir = opts.get('ca_cert_dir', args.ca_cert_dir)
        args.ca_cert_dir = cast(Optional[str], ca_cert_dir)
        if args.ca_cert_dir is None:
            args.ca_cert_dir = os.path.join(
                args.proxy_py_data_dir, 'certificates',
            )
        os.makedirs(args.ca_cert_dir, exist_ok=True)

        # FIXME: Necessary here until flags framework provides a way
        # for flag owners to initialize
        os.makedirs(args.cache_dir, exist_ok=True)
        os.makedirs(os.path.join(args.cache_dir, 'responses'), exist_ok=True)
        os.makedirs(os.path.join(args.cache_dir, 'content'), exist_ok=True)

        return args

    @staticmethod
    def get_default_plugins(
            args: argparse.Namespace,
    ) -> List[str]:
        """Prepare list of plugins to load based upon
        --enable-* and --disable-* flags.
        """
        default_plugins: List[str] = []
        if hasattr(args, 'enable_dashboard') and args.enable_dashboard:
            default_plugins.append(PLUGIN_WEB_SERVER)
            args.enable_static_server = True
            default_plugins.append(PLUGIN_DASHBOARD)
            default_plugins.append(PLUGIN_WEBSOCKET_TRANSPORT)
            default_plugins.append(PLUGIN_INSPECT_TRAFFIC)
            args.enable_events = True
            args.enable_devtools = True
        if hasattr(args, 'enable_metrics') and args.enable_metrics:
            default_plugins.append(PLUGIN_WEB_SERVER)
            default_plugins.append(PLUGIN_METRICS)
            args.enable_events = True
        if hasattr(args, 'enable_devtools') and args.enable_devtools:
            default_plugins.append(PLUGIN_DEVTOOLS_PROTOCOL)
            default_plugins.append(PLUGIN_WEB_SERVER)
        if not args.disable_http_proxy:
            default_plugins.append(PLUGIN_HTTP_PROXY)
        if args.enable_web_server or \
                args.pac_file is not None or \
                args.enable_static_server:
            default_plugins.append(PLUGIN_WEB_SERVER)
        if args.enable_reverse_proxy:
            default_plugins.append(PLUGIN_WEB_SERVER)
            default_plugins.append(PLUGIN_REVERSE_PROXY)
        if args.pac_file is not None:
            default_plugins.append(PLUGIN_PAC_FILE)
        return list(collections.OrderedDict.fromkeys(default_plugins).keys())


flags = FlagParser()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import asyncio
from abc import abstractmethod
from typing import Any, Optional
from multiprocessing import connection

from .threadless import Threadless


class BaseRemoteExecutor(Threadless[connection.Connection]):
    """A threadless executor implementation which receives work over a connection."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._loop: Optional[asyncio.AbstractEventLoop] = None

    @property
    def loop(self) -> Optional[asyncio.AbstractEventLoop]:
        if self._loop is None:
            self._loop = asyncio.get_event_loop_policy().get_event_loop()
        return self._loop

    def work_queue_fileno(self) -> Optional[int]:
        return self.work_queue.fileno()

    def close_work_queue(self) -> None:
        self.work_queue.close()

    def receive_from_work_queue(self) -> bool:
        self.work(self.work_queue.recv())
        return False

    @abstractmethod
    def work(self, *args: Any) -> None:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import logging
import selectors
from abc import abstractmethod
from typing import Any, Optional

from .tcp_server import BaseTcpServerHandler
from ..connection import TcpClientConnection, TcpServerConnection
from ...http.parser import HttpParser, httpParserTypes
from ...common.types import Readables, Writables, SelectableEvents
from ...common.utils import text_


logger = logging.getLogger(__name__)


class BaseTcpTunnelHandler(BaseTcpServerHandler[TcpClientConnection]):
    """BaseTcpTunnelHandler build on-top of BaseTcpServerHandler work class.

    On-top of BaseTcpServerHandler implementation,
    BaseTcpTunnelHandler introduces an upstream TcpServerConnection
    and adds it to the core event loop when needed.

    Currently, implementations must call connect_upstream from within
    handle_data.  See HttpsConnectTunnelHandler for example usage.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.request = HttpParser(
            httpParserTypes.REQUEST_PARSER,
            enable_proxy_protocol=self.flags.enable_proxy_protocol,
        )
        self.upstream: Optional[TcpServerConnection] = None

    @abstractmethod
    def handle_data(self, data: memoryview) -> Optional[bool]:
        pass    # pragma: no cover

    @staticmethod
    def create(*args: Any) -> TcpClientConnection:   # pragma: no cover
        return TcpClientConnection(*args)

    def initialize(self) -> None:
        self.work.connection.setblocking(False)

    def shutdown(self) -> None:
        if self.upstream:
            logger.debug(
                'Connection closed with upstream {0}:{1}'.format(
                    text_(self.request.host), self.request.port,
                ),
            )
            self.upstream.close()
        super().shutdown()

    async def get_events(self) -> SelectableEvents:
        # Get default client events
        ev: SelectableEvents = await super().get_events()
        # Read from server if we are connected
        if self.upstream and self.upstream._conn is not None:
            ev[self.upstream.connection.fileno()] = selectors.EVENT_READ
        # If there is pending buffer for server
        # also register for EVENT_WRITE events
        if self.upstream and self.upstream.has_buffer():
            if self.upstream.connection.fileno() in ev:
                ev[self.upstream.connection.fileno()] |= selectors.EVENT_WRITE
            else:
                ev[self.upstream.connection.fileno()] = selectors.EVENT_WRITE
        return ev

    async def handle_events(
            self,
            readables: Readables,
            writables: Writables,
    ) -> bool:
        # Handle client events
        do_shutdown: bool = await super().handle_events(readables, writables)
        if do_shutdown:
            return do_shutdown
        # Handle server events
        if self.upstream and self.upstream.connection.fileno() in readables:
            data = self.upstream.recv(self.flags.server_recvbuf_size)
            if data is None:
                # Server closed connection
                logger.debug('Connection closed by server')
                return True
            # tunnel data to client
            self.work.queue(data)
        if self.upstream and self.upstream.connection.fileno() in writables:
            self.upstream.flush(self.flags.max_sendbuf_size)
        return False

    def connect_upstream(self) -> None:
        assert self.request.host and self.request.port
        self.upstream = TcpServerConnection(
            text_(self.request.host), self.request.port,
        )
        self.upstream.connect()
        logger.debug(
            'Connection established with upstream {0}:{1}'.format(
                text_(self.request.host), self.request.port,
            ),
        )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       pac
"""
from typing import Any, List, Tuple, Optional

from .plugin import HttpWebServerBasePlugin
from ..parser import HttpParser
from .protocols import httpProtocolTypes
from ..responses import okResponse
from ...common.flag import flags
from ...common.utils import text_, bytes_
from ...common.constants import DEFAULT_PAC_FILE, DEFAULT_PAC_FILE_URL_PATH


flags.add_argument(
    '--pac-file',
    type=str,
    default=DEFAULT_PAC_FILE,
    help='A file (Proxy Auto Configuration) or string to serve when '
    'the server receives a direct file request. '
    'Using this option enables proxy.HttpWebServerPlugin.',
)
flags.add_argument(
    '--pac-file-url-path',
    type=str,
    default=text_(DEFAULT_PAC_FILE_URL_PATH),
    help='Default: %s. Web server path to serve the PAC file.' %
    text_(DEFAULT_PAC_FILE_URL_PATH),
)


class HttpWebServerPacFilePlugin(HttpWebServerBasePlugin):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.pac_file_response: Optional[memoryview] = None
        self.cache_pac_file_response()

    def routes(self) -> List[Tuple[int, str]]:
        if self.flags.pac_file_url_path:
            return [
                (
                    httpProtocolTypes.HTTP, r'{0}$'.format(
                        text_(self.flags.pac_file_url_path),
                    ),
                ),
                (
                    httpProtocolTypes.HTTPS, r'{0}$'.format(
                        text_(self.flags.pac_file_url_path),
                    ),
                ),
            ]
        return []   # pragma: no cover

    def handle_request(self, request: HttpParser) -> None:
        if self.flags.pac_file and self.pac_file_response:
            self.client.queue(self.pac_file_response)

    def cache_pac_file_response(self) -> None:
        if self.flags.pac_file:
            try:
                with open(self.flags.pac_file, 'rb') as f:
                    content = f.read()
            except IOError:
                content = bytes_(self.flags.pac_file)
            self.pac_file_response = okResponse(
                content=content,
                headers={
                    b'Content-Type': b'application/x-ns-proxy-autoconfig',
                },
                conn_close=True,
                compress=False,
            )



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import base64
import socket
import secrets
import selectors
from typing import Callable, Optional

from .frame import WebsocketFrame
from ..parser import HttpParser, httpParserTypes
from ...common.types import TcpOrTlsSocket
from ...common.utils import (
    text_, new_socket_connection, build_websocket_handshake_request,
)
from ...core.connection import TcpConnection, tcpConnectionTypes
from ...common.constants import (
    DEFAULT_BUFFER_SIZE, DEFAULT_SELECTOR_SELECT_TIMEOUT,
)


class WebsocketClient(TcpConnection):
    """Websocket client connection.

    TODO: Make me compatible with the work framework."""

    def __init__(
        self,
        hostname: bytes,
        port: int,
        path: bytes = b'/',
        on_message: Optional[Callable[[WebsocketFrame], None]] = None,
    ) -> None:
        super().__init__(tcpConnectionTypes.CLIENT)
        self.hostname: bytes = hostname
        self.port: int = port
        self.path: bytes = path
        self.sock: socket.socket = new_socket_connection(
            (socket.gethostbyname(text_(self.hostname)), self.port),
        )
        self.on_message: Optional[
            Callable[
                [
                    WebsocketFrame,
                ],
                None,
            ]
        ] = on_message
        self.selector: selectors.DefaultSelector = selectors.DefaultSelector()

    @property
    def connection(self) -> TcpOrTlsSocket:
        return self.sock

    def handshake(self) -> None:
        """Start websocket upgrade & handshake protocol"""
        self.upgrade()
        self.sock.setblocking(False)

    def upgrade(self) -> None:
        """Creates a key and sends websocket handshake packet to upstream.
        Receives response from the server and asserts that websocket
        accept header is valid in the response."""
        key = base64.b64encode(secrets.token_bytes(16))
        self.sock.send(
            build_websocket_handshake_request(
                key,
                url=self.path,
                host=self.hostname,
            ),
        )
        response = HttpParser(httpParserTypes.RESPONSE_PARSER)
        response.parse(memoryview(self.sock.recv(DEFAULT_BUFFER_SIZE)))
        accept = response.header(b'Sec-Websocket-Accept')
        assert WebsocketFrame.key_to_accept(key) == accept

    def shutdown(self, _data: Optional[bytes] = None) -> None:
        """Closes connection with the server."""
        super().close()

    def run_once(self) -> bool:
        ev = selectors.EVENT_READ
        if self.has_buffer():
            ev |= selectors.EVENT_WRITE
        self.selector.register(self.sock.fileno(), ev)
        events = self.selector.select(timeout=DEFAULT_SELECTOR_SELECT_TIMEOUT)
        self.selector.unregister(self.sock)
        for _, mask in events:
            if mask & selectors.EVENT_READ and self.on_message:
                # TODO: client recvbuf size flag currently not used here
                raw = self.recv()
                if raw is None or raw == b'':
                    self.closed = True
                    return True
                frame = WebsocketFrame()
                frame.parse(raw.tobytes())
                self.on_message(frame)
            elif mask & selectors.EVENT_WRITE:
                # TODO: max sendbuf size flag currently not used here
                self.flush()
        return False

    def run(self) -> None:
        try:
            while not self.closed:
                if self.run_once():
                    break
        except KeyboardInterrupt:
            pass
        finally:
            if not self.closed:
                self.selector.unregister(self.sock)
                try:
                    self.sock.shutdown(socket.SHUT_WR)
                except OSError:
                    pass
            self.sock.close()
            self.selector.close()



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       localhost
       httpbin
"""
from typing import Optional

from ..http.proxy import HttpProxyBasePlugin
from ..http.parser import HttpParser


class ModifyPostDataPlugin(HttpProxyBasePlugin):
    """Modify POST request body before sending to upstream server.

    Following curl executions will work:
        1. Plain
           curl -v -x localhost:8899 -X POST http://httpbin.org/post -d 'key=value'
        2. Chunked
           curl -v -x localhost:8899 -X POST \
               -H 'Transfer-Encoding: chunked' http://httpbin.org/post -d 'key=value'
        3. Chunked & Compressed
           echo 'key=value' | gzip | curl -v \
               -x localhost:8899 \
               -X POST \
               --data-binary @- -H 'Transfer-Encoding: chunked' \
               -H 'Content-Encoding: gzip' http://httpbin.org/post

    """

    MODIFIED_BODY = b'{"key": "modified"}'

    def before_upstream_connection(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        return request

    def handle_client_request(
            self, request: HttpParser,
    ) -> Optional[HttpParser]:
        if request.body:
            request.update_body(
                ModifyPostDataPlugin.MODIFIED_BODY,
                content_type=b'application/json',
            )
        return request



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import os
import ssl
import sys
import time
import pathlib
import secrets
import platform
import ipaddress
import sysconfig
from typing import Any, List

from .version import __version__


SYS_PLATFORM = platform.system()
IS_WINDOWS = SYS_PLATFORM.lower() in ('windows', 'cygwin')


def _env_threadless_compliant() -> bool:
    """Returns true for Python 3.8+ across all platforms
    except Windows."""
    return not IS_WINDOWS and sys.version_info >= (3, 8)


PROXY_PY_START_TIME = time.time()

# /path/to/proxy.py/proxy folder
PROXY_PY_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

# Path to virtualenv/lib/python3.X/site-packages
PROXY_PY_SITE_PACKAGES = sysconfig.get_path('purelib')
assert PROXY_PY_SITE_PACKAGES

CRLF = b'\r\n'
COLON = b':'
WHITESPACE = b' '
COMMA = b','
DOT = b'.'
SLASH = b'/'
AT = b'@'
AND = b'&'
EQUAL = b'='
TCP_PROTO = b"tcp"
UDP_PROTO = b"udp"
HTTP_PROTO = b'http'
HTTPS_PROTO = HTTP_PROTO + b's'
WEBSOCKET_PROTO = b"ws"
WEBSOCKETS_PROTO = WEBSOCKET_PROTO + b"s"
HTTP_PROTOS = [HTTP_PROTO, HTTPS_PROTO, WEBSOCKET_PROTO, WEBSOCKETS_PROTO]
SSL_PROTOS = [HTTPS_PROTO, WEBSOCKETS_PROTO]
HTTP_1_0 = HTTP_PROTO.upper() + SLASH + b'1.0'
HTTP_1_1 = HTTP_PROTO.upper() + SLASH + b'1.1'
COLON_SLASH_SLASH = COLON + SLASH + SLASH
TCP_URL_PREFIX = TCP_PROTO + COLON_SLASH_SLASH
UDP_URL_PREFIX = UDP_PROTO + COLON_SLASH_SLASH
HTTP_URL_PREFIX = HTTP_PROTO + COLON_SLASH_SLASH
HTTPS_URL_PREFIX = HTTPS_PROTO + COLON_SLASH_SLASH
WEBSOCKET_URL_PREFIX = WEBSOCKET_PROTO + COLON_SLASH_SLASH
WEBSOCKETS_URL_PREFIX = WEBSOCKETS_PROTO + COLON_SLASH_SLASH

LOCAL_INTERFACE_HOSTNAMES = (
    b'localhost',
    b'127.0.0.1',
    b'::1',
)

ANY_INTERFACE_HOSTNAMES = (
    b'0.0.0.0',
    b'::',
)

PROXY_AGENT_HEADER_KEY = b'Proxy-agent'
PROXY_AGENT_HEADER_VALUE = b'proxy.py v' + \
    __version__.encode('utf-8', 'strict')
PROXY_AGENT_HEADER = PROXY_AGENT_HEADER_KEY + \
    COLON + WHITESPACE + PROXY_AGENT_HEADER_VALUE

# Defaults
DEFAULT_BACKLOG = 100
DEFAULT_BASIC_AUTH = None
DEFAULT_MAX_SEND_SIZE = 64 * 1024
DEFAULT_BUFFER_SIZE = 128 * 1024
DEFAULT_CA_CERT_DIR = None
DEFAULT_INSECURE_TLS_INTERCEPTION = False
DEFAULT_CA_CERT_FILE = None
DEFAULT_CA_KEY_FILE = None
DEFAULT_CA_SIGNING_KEY_FILE = None
DEFAULT_CERT_FILE = None
DEFAULT_CA_FILE = pathlib.Path(
    PROXY_PY_SITE_PACKAGES,
) / 'certifi' / 'cacert.pem'
DEFAULT_CLIENT_RECVBUF_SIZE = DEFAULT_BUFFER_SIZE
DEFAULT_DEVTOOLS_WS_PATH = b'/devtools'
DEFAULT_DISABLE_HEADERS: List[bytes] = []
DEFAULT_DISABLE_HTTP_PROXY = False
DEFAULT_ENABLE_DASHBOARD = False
DEFAULT_ENABLE_SSH_TUNNEL = False
DEFAULT_ENABLE_DEVTOOLS = False
DEFAULT_ENABLE_EVENTS = False
DEFAULT_EVENTS_QUEUE = None
DEFAULT_ENABLE_STATIC_SERVER = False
DEFAULT_ENABLE_WEB_SERVER = False
DEFAULT_ENABLE_REVERSE_PROXY = False
DEFAULT_ALLOWED_URL_SCHEMES = [HTTP_PROTO, HTTPS_PROTO]
DEFAULT_IPV4_HOSTNAME = ipaddress.IPv4Address('127.0.0.1')
DEFAULT_IPV6_HOSTNAME = ipaddress.IPv6Address('::1')
DEFAULT_KEY_FILE = None
DEFAULT_LOG_FILE = None
DEFAULT_LOG_FORMAT = '%(asctime)s - pid:%(process)d [%(levelname)-.1s] %(module)s.%(funcName)s:%(lineno)d - %(message)s'
DEFAULT_LOG_LEVEL = 'INFO'
DEFAULT_WEB_ACCESS_LOG_FORMAT = '{client_ip}:{client_port} - ' \
    '{request_method} {request_path} - {request_ua} - {connection_time_ms}ms'
DEFAULT_HTTP_PROXY_ACCESS_LOG_FORMAT = '{client_ip}:{client_port} - ' + \
    '{request_method} {server_host}:{server_port}{request_path} - ' + \
    '{response_code} {response_reason} - {response_bytes} bytes - ' + \
    '{connection_time_ms}ms'
DEFAULT_HTTPS_PROXY_ACCESS_LOG_FORMAT = '{client_ip}:{client_port} - ' + \
    '{request_method} {server_host}:{server_port} - ' + \
    '{response_bytes} bytes - {connection_time_ms}ms'
DEFAULT_REVERSE_PROXY_ACCESS_LOG_FORMAT = (
    "{client_ip}:{client_port} - "
    + "{request_method} {request_path} - {request_ua} -> "
    + "{upstream_proxy_pass} - {connection_time_ms}ms"
)
DEFAULT_NUM_ACCEPTORS = 0
DEFAULT_NUM_WORKERS = 0
DEFAULT_OPEN_FILE_LIMIT = 1024
DEFAULT_PAC_FILE = None
DEFAULT_PAC_FILE_URL_PATH = b'/'
DEFAULT_ENABLE_METRICS = False
DEFAULT_METRICS_URL_PATH = b"/metrics"
DEFAULT_PID_FILE = None
DEFAULT_PORT_FILE = None
DEFAULT_PLUGINS: List[Any] = []
DEFAULT_PORT = 8899
DEFAULT_SERVER_RECVBUF_SIZE = DEFAULT_BUFFER_SIZE
DEFAULT_STATIC_SERVER_DIR = os.path.join(PROXY_PY_DIR, "public")
DEFAULT_MIN_COMPRESSION_LENGTH = 20  # In bytes
DEFAULT_THREADLESS = _env_threadless_compliant()
DEFAULT_LOCAL_EXECUTOR = True
DEFAULT_TIMEOUT = 10.0
DEFAULT_CONNECT_TIMEOUT = 10.0
DEFAULT_VERSION = False
DEFAULT_HTTP_PORT = 80
DEFAULT_HTTPS_PORT = 443
DEFAULT_WORK_KLASS = 'proxy.http.HttpProtocolHandler'
DEFAULT_SSH_LISTENER_KLASS = "proxy.core.ssh.listener.SshTunnelListener"
DEFAULT_ENABLE_PROXY_PROTOCOL = False
# 25 milliseconds to keep the loops hot
# Will consume ~0.3-0.6% CPU when idle.
DEFAULT_SELECTOR_SELECT_TIMEOUT = 25 / 1000
DEFAULT_WAIT_FOR_TASKS_TIMEOUT = 1 / 1000
DEFAULT_INACTIVE_CONN_CLEANUP_TIMEOUT = 1   # in seconds
DEFAULT_SSL_CONTEXT_OPTIONS = (
    ssl.OP_NO_COMPRESSION
    if sys.version_info >= (3, 10)
    else (ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1)
)
DEFAULT_ENABLE_REWRITE_HOST = False

DEFAULT_DEVTOOLS_DOC_URL = 'http://proxy'
DEFAULT_DEVTOOLS_FRAME_ID = secrets.token_hex(8)
DEFAULT_DEVTOOLS_LOADER_ID = secrets.token_hex(8)

DEFAULT_DATA_DIRECTORY_PATH = os.path.join(str(pathlib.Path.home()), '.proxy')
DEFAULT_CACHE_DIRECTORY_PATH = os.path.join(
    DEFAULT_DATA_DIRECTORY_PATH, 'cache',
)
DEFAULT_CACHE_REQUESTS = False
DEFAULT_CACHE_BY_CONTENT_TYPE = False
DEFAULT_METRICS_DIRECTORY_PATH = os.path.join(DEFAULT_DATA_DIRECTORY_PATH, "metrics")

# Cor plugins enabled by default or via flags
DEFAULT_ABC_PLUGINS = [
    "HttpProtocolHandlerPlugin",
    "HttpProxyBasePlugin",
    "HttpWebServerBasePlugin",
    "WebSocketTransportBasePlugin",
    "ReverseProxyBasePlugin",
    "GroutClientBasePlugin",
]
PLUGIN_DASHBOARD = 'proxy.dashboard.ProxyDashboard'
PLUGIN_HTTP_PROXY = 'proxy.http.proxy.HttpProxyPlugin'
PLUGIN_PROXY_AUTH = 'proxy.http.proxy.auth.AuthPlugin'
PLUGIN_WEB_SERVER = 'proxy.http.server.HttpWebServerPlugin'
PLUGIN_REVERSE_PROXY = 'proxy.http.server.reverse.ReverseProxy'
PLUGIN_PAC_FILE = 'proxy.http.server.HttpWebServerPacFilePlugin'
PLUGIN_DEVTOOLS_PROTOCOL = 'proxy.http.inspector.devtools.DevtoolsProtocolPlugin'
PLUGIN_INSPECT_TRAFFIC = 'proxy.http.inspector.inspect_traffic.InspectTrafficPlugin'
PLUGIN_WEBSOCKET_TRANSPORT = 'proxy.http.websocket.transport.WebSocketTransport'
PLUGIN_METRICS = "proxy.http.server.MetricsWebServerPlugin"

PY2_DEPRECATION_MESSAGE = '''DEPRECATION: proxy.py no longer supports Python 2.7.  Kindly upgrade to Python 3+. '
                'If for some reasons you cannot upgrade, use'
                '"pip install proxy.py==0.3".'''



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import argparse
import itertools
from typing import TYPE_CHECKING, Any, List, Type

from .tcp import TcpSocketListener
from .unix import UnixSocketListener


if TYPE_CHECKING:   # pragma: no cover
    from .base import BaseListener


class ListenerPool:
    """Provides abstraction around starting multiple listeners
    based upon flags."""

    def __init__(self, flags: argparse.Namespace) -> None:
        self.flags = flags
        self.pool: List['BaseListener'] = []

    def __enter__(self) -> 'ListenerPool':
        self.setup()
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def setup(self) -> None:
        if self.flags.unix_socket_path:
            self.add(UnixSocketListener)
        hostnames = {self.flags.hostname, *self.flags.hostnames}
        ports = list(self.flags.ports)
        if not self.flags.unix_socket_path:
            ports.append(self.flags.port)
        for hostname, port in itertools.product(hostnames, ports):
            self.add(TcpSocketListener, hostname=hostname, port=port)

    def shutdown(self) -> None:
        for listener in self.pool:
            listener.shutdown()
        self.pool.clear()

    def add(self, klass: Type['BaseListener'], **kwargs: Any) -> None:
        listener = klass(flags=self.flags, **kwargs)
        listener.setup()
        self.pool.append(listener)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from ..common.types import Readables, Writables, Descriptors


# Since 3.4.0
class DescriptorsHandlerMixin:
    """DescriptorsHandlerMixin provides abstraction used by several core HTTP modules
    include web and proxy plugins.  By using DescriptorsHandlerMixin, class
    becomes complaint with core event loop."""

    # @abstractmethod
    async def get_descriptors(self) -> Descriptors:
        """Implementations must return a list of descriptions that they wish to
        read from and write into."""
        return [], []  # pragma: no cover

    # @abstractmethod
    async def write_to_descriptors(self, w: Writables) -> bool:
        """Implementations must now write/flush data over the socket.

        Note that buffer management is in-build into the connection classes.
        Hence implementations MUST call
        :meth:`~proxy.core.connection.connection.TcpConnection.flush`
        here, to send any buffered data over the socket.
        """
        return False  # pragma: no cover

    # @abstractmethod
    async def read_from_descriptors(self, r: Readables) -> bool:
        """Implementations must now read data over the socket."""
        return False  # pragma: no cover



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import re
import random
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Optional

from proxy.http import Url
from proxy.core.base import TcpUpstreamConnectionHandler
from proxy.http.parser import HttpParser
from proxy.http.server import HttpWebServerBasePlugin
from proxy.common.utils import text_, bytes_
from proxy.http.exception import HttpProtocolException
from proxy.common.constants import (
    COLON, HTTP_PROTO, HTTPS_PROTO, DEFAULT_HTTP_PORT, DEFAULT_HTTPS_PORT,
    DEFAULT_REVERSE_PROXY_ACCESS_LOG_FORMAT,
)
from ...common.types import Readables, Writables, Descriptors


if TYPE_CHECKING:   # pragma: no cover
    from .plugin import ReverseProxyBasePlugin


logger = logging.getLogger(__name__)


class ReverseProxy(TcpUpstreamConnectionHandler, HttpWebServerBasePlugin):
    """Extend in-built Web Server to add Reverse Proxy capabilities."""

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.choice: Optional[Url] = None
        self.plugins: List['ReverseProxyBasePlugin'] = []
        for klass in self.flags.plugins[b'ReverseProxyBasePlugin']:
            plugin: 'ReverseProxyBasePlugin' = klass(
                self.uid, self.flags, self.client, self.event_queue, self.upstream_conn_pool,
            )
            self.plugins.append(plugin)
        self._upstream_proxy_pass: Optional[str] = None
        self._needs_upstream: bool = False

    def do_upgrade(self, request: HttpParser) -> bool:
        """Signal web protocol handler to not upgrade websocket requests by default."""
        return False

    def handle_upstream_data(self, raw: memoryview) -> None:
        # TODO: Parse response and implement plugin hook per parsed response object
        # This will give plugins a chance to modify the responses before dispatching to client
        self.client.queue(raw)

    def routes(self) -> List[Tuple[int, str]]:
        r = []
        for plugin in self.plugins:
            for route in plugin.regexes():
                for proto in plugin.protocols():
                    r.append((proto, route))
        return r

    def handle_request(self, request: HttpParser) -> None:
        # before_routing
        for plugin in self.plugins:
            r = plugin.before_routing(request)
            if r is None:
                raise HttpProtocolException('before_routing closed connection')
            request = r

        # routes
        for plugin in self.plugins:
            for route in plugin.routes():
                # Static routes
                if isinstance(route, tuple):
                    pattern = re.compile(route[0])
                    if pattern.match(text_(request.path)):
                        self.choice = Url.from_bytes(
                            random.choice(route[1]),
                        )
                        self._needs_upstream = True
                        break
                # Dynamic routes
                elif isinstance(route, str):
                    pattern = re.compile(route)
                    if pattern.match(text_(request.path)):
                        choice = plugin.handle_route(request, pattern)
                        if isinstance(choice, Url):
                            self.choice = choice
                            self._needs_upstream = True
                            self._upstream_proxy_pass = str(self.choice)
                        elif isinstance(choice, memoryview):
                            self.client.queue(choice)
                            self._upstream_proxy_pass = '{0} bytes'.format(len(choice))
                        else:
                            self.upstream = choice
                            self._upstream_proxy_pass = '{0}:{1}'.format(
                                *self.upstream.addr,
                            )
                        break
                else:
                    raise ValueError('Invalid route')

        if self._needs_upstream:
            assert self.choice and self.choice.hostname
            port = (
                self.choice.port or DEFAULT_HTTP_PORT
                if self.choice.scheme == HTTP_PROTO
                else self.choice.port or DEFAULT_HTTPS_PORT
            )
            self.initialize_upstream(text_(self.choice.hostname), port)
            assert self.upstream
            try:
                self.upstream.connect()
                if self.choice.scheme == HTTPS_PROTO:
                    self.upstream.wrap(
                        text_(self.choice.hostname),
                        as_non_blocking=True,
                        ca_file=self.flags.ca_file,
                    )
                request.path = self.choice.remainder
                self.upstream.queue(
                    memoryview(
                        request.build(
                            host=(
                                self.choice.hostname
                                + (
                                    COLON + bytes_(self.choice.port)
                                    if self.choice.port is not None
                                    else b''
                                )
                                if self.flags.rewrite_host_header
                                else None
                            ),
                        ),
                    ),
                )
            except ConnectionRefusedError:
                raise HttpProtocolException(  # pragma: no cover
                    'Connection refused by upstream server {0}:{1}'.format(
                        text_(self.choice.hostname),
                        port,
                    ),
                )

    def on_client_connection_close(self) -> None:
        if self._needs_upstream and self.upstream and not self.upstream.closed:
            logger.debug('Closing upstream server connection')
            self.upstream.close()
        self.upstream = None

    def on_client_data(
        self,
        request: HttpParser,
        raw: memoryview,
    ) -> Optional[memoryview]:
        if request.is_websocket_upgrade:
            assert self.upstream
            self.upstream.queue(raw)
        return raw

    def on_access_log(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        context.update(
            {
                'upstream_proxy_pass': self._upstream_proxy_pass,
            },
        )
        log_handled = False
        for plugin in self.plugins:
            ctx = plugin.on_access_log(context)
            if ctx is None:
                log_handled = True
                break
            context = ctx
        if not log_handled:
            logger.info(DEFAULT_REVERSE_PROXY_ACCESS_LOG_FORMAT.format_map(context))
        return None

    async def get_descriptors(self) -> Descriptors:
        r, w = await super().get_descriptors()
        # TODO(abhinavsingh): We need to keep a mapping of plugin and
        # descriptors registered by them, so that within write/read blocks
        # we can invoke the right plugin callbacks.
        for plugin in self.plugins:
            plugin_read_desc, plugin_write_desc = await plugin.get_descriptors()
            r.extend(plugin_read_desc)
            w.extend(plugin_write_desc)
        return r, w

    async def read_from_descriptors(self, r: Readables) -> bool:
        for plugin in self.plugins:
            teardown = await plugin.read_from_descriptors(r)
            if teardown:
                return True
        return await super().read_from_descriptors(r)

    async def write_to_descriptors(self, w: Writables) -> bool:
        for plugin in self.plugins:
            teardown = await plugin.write_to_descriptors(w)
            if teardown:
                return True
        return await super().write_to_descriptors(w)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from .test_case import TestCase


__all__ = [
    'TestCase',
]



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       iterable
"""
from typing import NamedTuple


# Ref: https://www.iana.org/assignments/http-methods/http-methods.xhtml
HttpHeaders = NamedTuple(
    'HttpHeaders', [
        ('PROXY_AUTHORIZATION', bytes),
        ('PROXY_CONNECTION', bytes),
    ],
)

httpHeaders = HttpHeaders(
    b'proxy-authorization',
    b'proxy-connection',
)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.

    .. spelling::

       http
       iterable
       websocket
       Websocket
"""
import io
import base64
import struct
import hashlib
import logging
import secrets
from typing import Type, TypeVar, Optional, NamedTuple


WebsocketOpcodes = NamedTuple(
    'WebsocketOpcodes', [
        ('CONTINUATION_FRAME', int),
        ('TEXT_FRAME', int),
        ('BINARY_FRAME', int),
        ('CONNECTION_CLOSE', int),
        ('PING', int),
        ('PONG', int),
    ],
)
websocketOpcodes = WebsocketOpcodes(0x0, 0x1, 0x2, 0x8, 0x9, 0xA)


V = TypeVar('V', bound='WebsocketFrame')

logger = logging.getLogger(__name__)


class WebsocketFrame:
    """Websocket frames parser and constructor."""

    GUID = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

    def __init__(self) -> None:
        self.fin: bool = False
        self.rsv1: bool = False
        self.rsv2: bool = False
        self.rsv3: bool = False
        self.opcode: int = 0
        self.masked: bool = False
        self.payload_length: Optional[int] = None
        self.mask: Optional[bytes] = None
        self.data: Optional[bytes] = None

    @classmethod
    def text(cls: Type[V], data: bytes) -> bytes:
        frame = cls()
        frame.fin = True
        frame.opcode = websocketOpcodes.TEXT_FRAME
        frame.data = data
        return frame.build()

    def reset(self) -> None:
        self.fin = False
        self.rsv1 = False
        self.rsv2 = False
        self.rsv3 = False
        self.opcode = 0
        self.masked = False
        self.payload_length = None
        self.mask = None
        self.data = None

    def parse_fin_and_rsv(self, byte: int) -> None:
        self.fin = bool(byte & 1 << 7)
        self.rsv1 = bool(byte & 1 << 6)
        self.rsv2 = bool(byte & 1 << 5)
        self.rsv3 = bool(byte & 1 << 4)
        self.opcode = byte & 0b00001111

    def parse_mask_and_payload(self, byte: int) -> None:
        self.masked = bool(byte & 0b10000000)
        self.payload_length = byte & 0b01111111

    def build(self) -> bytes:
        if self.payload_length is None and self.data:
            self.payload_length = len(self.data)
        raw = io.BytesIO()
        raw.write(
            struct.pack(
                '!B',
                (1 << 7 if self.fin else 0) |
                (1 << 6 if self.rsv1 else 0) |
                (1 << 5 if self.rsv2 else 0) |
                (1 << 4 if self.rsv3 else 0) |
                self.opcode,
            ),
        )
        assert self.payload_length is not None
        if self.payload_length < 126:
            raw.write(
                struct.pack(
                    '!B',
                    (1 << 7 if self.masked else 0) | self.payload_length,
                ),
            )
        elif self.payload_length < 1 << 16:
            raw.write(
                struct.pack(
                    '!BH',
                    (1 << 7 if self.masked else 0) | 126,
                    self.payload_length,
                ),
            )
        elif self.payload_length < 1 << 64:
            raw.write(
                struct.pack(
                    '!BHQ',
                    (1 << 7 if self.masked else 0) | 127,
                    self.payload_length,
                ),
            )
        else:
            raise ValueError(
                f'Invalid payload_length { self.payload_length},'
                f'maximum allowed { 1 << 64}',
            )
        if self.masked and self.data:
            mask = secrets.token_bytes(4) if self.mask is None else self.mask
            raw.write(mask)
            raw.write(self.apply_mask(self.data, mask))
        elif self.data:
            raw.write(self.data)
        return raw.getvalue()

    def parse(self, raw: bytes) -> bytes:
        cur = 0
        self.parse_fin_and_rsv(raw[cur])
        cur += 1

        self.parse_mask_and_payload(raw[cur])
        cur += 1

        if self.payload_length == 126:
            data = raw[cur: cur + 2]
            self.payload_length, = struct.unpack('!H', data)
            cur += 2
        elif self.payload_length == 127:
            data = raw[cur: cur + 8]
            self.payload_length, = struct.unpack('!Q', data)
            cur += 8

        if self.masked:
            self.mask = raw[cur: cur + 4]
            cur += 4

        assert self.payload_length
        self.data = raw[cur: cur + self.payload_length]
        cur += self.payload_length
        if self.masked:
            assert self.mask is not None
            self.data = self.apply_mask(self.data, self.mask)

        return raw[cur:]

    @staticmethod
    def apply_mask(data: bytes, mask: bytes) -> bytes:
        raw = bytearray(data)
        for i in range(len(raw)):
            raw[i] = raw[i] ^ mask[i % 4]
        return bytes(raw)

    @staticmethod
    def key_to_accept(key: bytes) -> bytes:
        sha1 = hashlib.sha1()   # noqa: S324
        sha1.update(key + WebsocketFrame.GUID)
        return base64.b64encode(sha1.digest())



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
from typing import NamedTuple


TlsContentType = NamedTuple(
    'TlsContentType', [
        ('CHANGE_CIPHER_SPEC', int),
        ('ALERT', int),
        ('HANDSHAKE', int),
        ('APPLICATION_DATA', int),
        ('OTHER', int),
    ],
)
tlsContentType = TlsContentType(20, 21, 22, 23, 255)


TlsHandshakeType = NamedTuple(
    'TlsHandshakeType', [
        ('HELLO_REQUEST', int),
        ('CLIENT_HELLO', int),
        ('SERVER_HELLO', int),
        ('CERTIFICATE', int),
        ('SERVER_KEY_EXCHANGE', int),
        ('CERTIFICATE_REQUEST', int),
        ('SERVER_HELLO_DONE', int),
        ('CERTIFICATE_VERIFY', int),
        ('CLIENT_KEY_EXCHANGE', int),
        ('FINISHED', int),
        ('OTHER', int),
    ],
)
tlsHandshakeType = TlsHandshakeType(0, 1, 2, 11, 12, 13, 14, 15, 16, 20, 255)



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import struct
import logging
from typing import Tuple, Optional

from .types import tlsContentType
from .handshake import TlsHandshake
from .certificate import TlsCertificate


logger = logging.getLogger(__name__)


class TlsParser:
    """TLS packet parser"""

    def __init__(self) -> None:
        self.content_type: int = tlsContentType.OTHER
        self.protocol_version: Optional[bytes] = None
        self.length: Optional[bytes] = None
        # only parse hand shake payload temporary
        self.handshake: Optional[TlsHandshake] = None
        self.certificate: Optional[TlsCertificate]

    def parse(self, raw: bytes) -> Tuple[bool, bytes]:
        """Parse TLS fragmentation.

        References

        https://datatracker.ietf.org/doc/html/rfc5246#page-15
        https://datatracker.ietf.org/doc/html/rfc5077#page-3
        https://datatracker.ietf.org/doc/html/rfc8446#page-10
        """
        length = len(raw)
        if length < 5:
            logger.debug('invalid data, len(raw) = %s', length)
            return False, raw
        payload_length, = struct.unpack('!H', raw[3:5])
        if length < 5 + payload_length:
            logger.debug(
                'incomplete data, len(raw) = %s, len(payload) = %s', length, payload_length,
            )
            return False, raw
        # parse
        self.content_type = raw[0]
        self.protocol_version = raw[1:3]
        self.length = raw[3:5]
        payload = raw[5:5 + payload_length]
        if self.content_type == tlsContentType.HANDSHAKE:
            # parse handshake
            self.handshake = TlsHandshake()
            self.handshake.parse(payload)
        return True, raw[5 + payload_length:]

    def build(self) -> bytes:
        data = b''
        data += bytes([self.content_type])
        assert self.protocol_version
        data += self.protocol_version
        payload = b''
        if self.content_type == tlsContentType.HANDSHAKE:
            assert self.handshake
            payload += self.handshake.build()
        length = struct.pack('!H', len(payload))
        data += length
        data += payload
        return data



# -*- coding: utf-8 -*-
"""
    proxy.py
    ~~~~~~~~
    ⚡⚡⚡ Fast, Lightweight, Pluggable, TLS interception capable proxy server focused on
    Network monitoring, controls & Application development, testing, debugging.

    :copyright: (c) 2013-present by Abhinav Singh and contributors.
    :license: BSD, see LICENSE for more details.
"""
import argparse
import mimetypes
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union, Optional

from ..parser import HttpParser
from ...http.url import Url
from ..responses import NOT_FOUND_RESPONSE_PKT, okResponse
from ..websocket import WebsocketFrame
from ..connection import HttpClientConnection
from ...core.event import EventQueue
from ..descriptors import DescriptorsHandlerMixin
from ...common.types import RePattern
from ...common.utils import bytes_
from ...http.server.protocols import httpProtocolTypes


if TYPE_CHECKING:   # pragma: no cover
    from ...core.connection import TcpServerConnection, UpstreamConnectionPool


class HttpWebServerBasePlugin(DescriptorsHandlerMixin, ABC):
    """Web Server Plugin for routing of requests."""

    def __init__(
            self,
            uid: str,
            flags: argparse.Namespace,
            client: HttpClientConnection,
            event_queue: EventQueue,
            upstream_conn_pool: Optional['UpstreamConnectionPool'] = None,
    ):
        self.uid = uid
        self.flags = flags
        self.client = client
        self.event_queue = event_queue
        self.upstream_conn_pool = upstream_conn_pool

    @staticmethod
    def serve_static_file(
        path: str,
        min_compression_length: int,
        compress: bool = True,
    ) -> memoryview:
        try:
            with open(path, 'rb') as f:
                content = f.read()
            content_type = mimetypes.guess_type(path)[0]
            if content_type is None:
                content_type = 'text/plain'
            headers = {
                b'Content-Type': bytes_(content_type),
                b'Cache-Control': b'max-age=86400',
            }
            return okResponse(
                content=content,
                headers=headers,
                compress=compress,
                min_compression_length=min_compression_length,
                # TODO: Should we really close or take advantage of keep-alive?
                conn_close=True,
            )
        except OSError:
            return NOT_FOUND_RESPONSE_PKT

    def name(self) -> str:
        """A unique name for your plugin.

        Defaults to name of the class. This helps plugin developers to directly
        access a specific plugin by its name."""
        return self.__class__.__qualname__      # pragma: no cover

    @abstractmethod
    def routes(self) -> List[Tuple[int, str]]:
        """Return List(protocol, path) that this plugin handles."""
        raise NotImplementedError()     # pragma: no cover

    @abstractmethod
    def handle_request(self, request: HttpParser) -> None:
        """Handle the request and serve response."""
        raise NotImplementedError()     # pragma: no cover

    def on_client_connection_close(self) -> None:
        """Client has closed the connection, do any clean up task now."""
        pass

    def do_upgrade(self, request: HttpParser) -> bool:
        return True

    def on_client_data(
        self,
        request: HttpParser,
        raw: memoryview,
    ) -> Optional[memoryview]:
        """Return None to avoid default webserver parsing of client data."""
        return raw

    # No longer abstract since v2.4.0
    #
    # @abstractmethod
    def on_websocket_open(self) -> None:
        """Called when websocket handshake has finished."""
        pass        # pragma: no cover

    # No longer abstract since v2.4.0
    #
    # @abstractmethod
    def on_websocket_message(self, frame: WebsocketFrame) -> None:
        """Handle websocket frame."""
        return None     # pragma: no cover

    # Deprecated since v2.4.0
    #
    # Instead use on_client_connection_close.
    #
    # This callback is no longer invoked.  Kindly
    # update your plugin before upgrading to v2.4.0.
    #
    # @abstractmethod
    # def on_websocket_close(self) -> None:
    #     """Called when websocket connection has been closed."""
    #     raise NotImplementedError()     # pragma: no cover

    def on_access_log(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """Use this method to override default access log format (see
        DEFAULT_WEB_ACCESS_LOG_FORMAT) or to add/update/modify passed context
        for usage by default access logger.

        Return updated log context to use for default logging format, OR
        Return None if plugin has logged the request.
        """
        return context


class ReverseProxyBasePlugin(DescriptorsHandlerMixin, ABC):
    """ReverseProxy base plugin class."""

    def __init__(
            self,
            uid: str,
            flags: argparse.Namespace,
            client: HttpClientConnection,
            event_queue: EventQueue,
            upstream_conn_pool: Optional['UpstreamConnectionPool'] = None,
    ):
        self.uid = uid
        self.flags = flags
        self.client = client
        self.event_queue = event_queue
        self.upstream_conn_pool = upstream_conn_pool

    @abstractmethod
    def routes(self) -> List[Union[str, Tuple[str, List[bytes]]]]:
        """List of routes registered by plugin.

        There are 2 types of routes:

        1) Dynamic routes (str): Should be a regular expression
        2) Static routes (tuple): Contain 2 elements, a route regular expression
           and list of upstream urls to serve when the route matches.

        Static routes doesn't require you to implement the `handle_route` method.
        Reverse proxy core will automatically pick one of the configured upstream URL
        and serve it out-of-box.

        Dynamic routes are helpful when you want to dynamically match and serve upstream urls.
        To handle dynamic routes, you must implement the `handle_route` method, which
        must return the url to serve."""
        raise NotImplementedError()     # pragma: no cover

    def protocols(self) -> List[int]:
        return [
            httpProtocolTypes.HTTP,
            httpProtocolTypes.HTTPS,
            httpProtocolTypes.WEBSOCKET,
        ]

    def before_routing(self, request: HttpParser) -> Optional[HttpParser]:
        """Plugins can modify request, return response, close connection.

        If None is returned, request will be dropped and closed."""
        return request  # pragma: no cover

    def handle_route(
        self,
        request: HttpParser,
        pattern: RePattern,
    ) -> Union[memoryview, Url, 'TcpServerConnection']:
        """Implement this method if you have configured dynamic routes."""
        raise NotImplementedError()

    def regexes(self) -> List[str]:
        """Helper method to return list of route regular expressions."""
        routes = []
        for route in self.routes():
            if isinstance(route, str):
                routes.append(route)
            elif isinstance(route, tuple):
                routes.append(route[0])
            else:
                raise ValueError('Invalid route type')
        return routes

    def on_access_log(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        """Use this method to override default access log format (see
        DEFAULT_REVERSE_PROXY_ACCESS_LOG_FORMAT) or to add/update/modify passed context
        for usage by default access logger.

        Return updated log context to use for default logging format, OR
        Return None if plugin has logged the request.
        """
        return context
